# This script demonstrates plotting DALYs and DALY-related quantities for the
# reference implementation in the TB model

import atomica as at
import numpy as np

P = at.demo("tb")

d = at.PlotData(P.results[0], outputs="daly_rate", pops="total", accumulate="integrate", project=P)
figs = at.plot_series(d, axis="pops")

d = at.PlotData(P.results[0], outputs=["yll_rate", "yld_rate", "daly_rate"], pops="total", t_bins="all")
d.outputs["yll_rate"] = "Years lost to deaths"
d.outputs["yld_rate"] = "Years lost to disease"
d.outputs["daly_rate"] = "DALYs"
figs = at.plot_bars(d, stack_outputs=[["yll_rate", "yld_rate"]])

d = at.PlotData(P.results[0], outputs=["yld_rate", "ac_inf"], pops="total", accumulate="integrate", project=P)
figs = at.plot_series(d, axis="outputs")



# This file produces calibrations of each YAML file shown in T7: YAML autocalibration tutorial.
# Output files will be saved locally in the outputs folder.
# Desirable results must be manually moved to the calibrations folder so that they can be plotted by the
# YAML_tutorial_plots_final.ipynb notebook. This is to avoid crowding the final results folder (calibrations) when
# re-running the calibrations or trying out new versions of the model.

import sys

in_path = "../inputs"
sys.path.append(in_path)
import atomica as at
import atomica.yaml_calibration
import os
from os.path import isfile
import time
import shutil
import winsound

yaml_dir = os.getcwd()
fnames = [f for f in os.listdir(yaml_dir) if isfile(f"{yaml_dir}/{f}")]
yaml_files = [f for f in fnames if ".yaml" in f and "T7_YAML" in f]
yaml_files += "WE_8"

testing = False  # <------------
savecals = True
out_path = "outputs"

# make project   ------------------------------------------------
inputs = "../assets"
F = at.ProjectFramework(f"{inputs}/T7_framework.xlsx")
D = at.ProjectData.from_spreadsheet(f"{inputs}/T7_databook.xlsx", framework=F)

P = at.Project(framework=F, databook=D, do_run=False)
P.settings.update_time_vector(start=2000, end=2050, dt=1 / 52)
cal = P.make_parset()
# starting calibration (if needed)
# cal.load_calibration(out_path + "/PAK_typh_YAML_autocalibrations_2024-04-25_1941_reinitialising/PAK_typh_YAML-autocalibration_2024-04-25_1959.xlsx")
# print('Existing calibration loaded')

for yaml_fname in yaml_files:
    yaml_path = f"{yaml_dir}/{yaml_fname}"

    # new folder to put calibrations in
    date = time.strftime("%Y-%m-%d_%H%M")
    calspath = f"{out_path}/cals_{yaml_fname[8]}"
    if savecals:
        # save everything in folder
        try:
            os.makedirs(calspath)
        except FileExistsError:
            calspath += f"_{date}"
            os.makedirs(f"{calspath}")

        # save yaml file for reference
        new_fname = f"{yaml_fname}_{date}.yaml"
        shutil.copy(yaml_path, calspath)
        shutil.move(f"{calspath}/{yaml_fname}", f"{calspath}/{new_fname}")  # renaming file

        # save logfile
        at.start_logging(f"{calspath}/logfile_{date}.txt")

        # save current runfile for reference
        runfile_name = f"{os.path.basename(__file__)}_{date}"
        shutil.copy(__file__, f"{calspath}/{runfile_name}")

    # display tree
    cal_tree = at.yaml_calibration.build(yaml_path)
    print(cal_tree)

    # run calibration w yaml instructions # <--------------------------
    newcal = P.calibrate(parset=cal, yaml=yaml_path, quiet=False, savedir=calspath, save_intermediate=False, verbose=0)

    # save final calibration
    newcal.save_calibration(f"{calspath}/cal_{yaml_fname[8]}")  # newcal.save_calibration(f'{calspath}/cal_{yaml_fname}')
    at.stop_logging()


# Notification - calibration finished
winsound.Beep(frequency=2500, duration=200)
winsound.Beep(frequency=2750, duration=200)
winsound.Beep(frequency=3050, duration=200)



import atomica as at
from atomica import ProjectFramework
import sciris as sc
from atomica import InvalidCascade
import os
import pytest

testdir = at.parent_dir()
tmpdir = testdir / "temp"

try:
    os.makedirs(tmpdir)
except FileExistsError:
    pass


def test_cascade_validate():
    from atomica.cascade import validate_cascade

    # First check that all of the library frameworks are OK
    fnames = os.listdir(at.LIBRARY_PATH)
    # NB. To test a single file, set e.g. `fnames=['framework_tb.xlsx']`
    for fname in fnames:
        if fname.endswith("_framework.xlsx") and not fname.startswith("~"):
            print("Validating %s" % (fname))
            F = ProjectFramework(at.LIBRARY_PATH / fname)

            # Validate all of the cascades in the framework
            for cascade in F.cascades:
                validate_cascade(F, cascade)

    for fname in ["framework_sir_badcascade1.xlsx", "framework_sir_badcascade2.xlsx"]:
        with pytest.raises(InvalidCascade):
            ProjectFramework(at.parent_dir() / fname)


def test_cascade_basic_tb():
    P = at.demo("tb")
    result = P.results[-1]

    result.plot(plot_name="Active DS-TB", project=P)

    result.plot(plot_name="Active DS-TB", pops="all", project=P)
    result.plot(plot_group="latency")

    #    # Export limited set of results based on 'Export' column in Framework, or export everything
    at.export_results(result, filename=tmpdir / "export_from_framework_1.xlsx")
    result.export_raw(filename=tmpdir / "export_raw.xlsx")  # Export everything

    # Plot various cascades
    startyear = 2000
    endyear = 2030

    at.plot_cascade(result, cascade="TB treatment (including recovered)", pops="all", year=startyear, data=P.data)
    at.plot_cascade(result, cascade="TB treatment (including recovered)", pops="all", year=endyear, data=P.data)

    at.plot_cascade(result, cascade="TB treatment (including recovered)", pops="0-4", year=endyear, data=P.data)
    at.plot_cascade(result, cascade="SP treatment", pops="0-4", year=endyear, data=P.data)

    at.plot_cascade(result, cascade="SP treatment", pops="Children 5-14", year=endyear, data=P.data)  # Look up using full name
    at.plot_cascade(result, cascade="SP treatment", pops=["Children 0-4", "Children 5-14"], year=endyear, data=P.data)  # Combine subset of pops - should be able to add numbers from the previous two figures


def test_cascade_basic_udt():
    P = at.demo("udt")
    result = P.results[-1]
    at.plot_cascade(result, pops="all", year=2016, data=P.data)  # plot default cascade


def test_cascade_scen_tb():

    P = at.demo("tb")
    par_results = P.results[-1]

    scvalues = dict()
    scen_par = "doth_rate"
    scen_pop = "0-4"

    scvalues[scen_par] = dict()
    scvalues[scen_par][scen_pop] = dict()
    scvalues[scen_par][scen_pop]["y"] = [0.5, 0.5]
    scvalues[scen_par][scen_pop]["t"] = [1999.0, 2050.0]
    scen = P.make_scenario(which="parameter", name="Increased deaths", scenario_values=scvalues)
    scen_results = scen.run(P, P.parsets["default"])

    par_results.name = "Baseline"
    scen_results.name = "Scenario"
    startyear = 2018
    endyear = 2020

    at.plot_multi_cascade([par_results, scen_results], None, year=startyear)
    at.plot_multi_cascade([par_results], None, year=[startyear, endyear])
    at.plot_multi_cascade([par_results, scen_results], cascade=0, pops="all", year=[startyear, endyear])
    at.plot_multi_cascade([par_results], cascade=1, year=[startyear, endyear])


def test_cascade_scen_udt():
    P = at.demo("udt")
    par_results = P.results[-1]
    scvalues = dict()

    scen_par = "num_diag"
    scen_pop = "adults"

    scvalues[scen_par] = dict()
    scvalues[scen_par][scen_pop] = dict()
    scvalues[scen_par][scen_pop]["y"] = [1000.0, 1500.0]
    scvalues[scen_par][scen_pop]["t"] = [2016.0, 2017.0]
    scen = P.make_scenario(which="parameter", name="Increased diagnosis rate", scenario_values=scvalues)
    scen_results = scen.run(P, P.parsets["default"])

    par_results.name = "Baseline"
    scen_results.name = "Scenario"
    startyear = 2016
    endyear = 2017

    at.plot_multi_cascade([par_results, scen_results], None, year=startyear)
    at.plot_multi_cascade([par_results], None, year=[startyear, endyear])


def test_cascade_dynamic():
    # Dynamically create a cascade
    P = at.demo("sir")
    cascade = sc.odict()
    cascade["Everyone"] = ["sus", "inf", "rec"]
    cascade["Ever infected"] = ["inf", "rec"]
    cascade["Recovered"] = "rec"
    at.plot_cascade(P.results[-1], cascade=cascade, pops="all", year=2023)


def test_cascade_sir():

    # Get a Result
    F = ProjectFramework(testdir / "framework_sir_dynamic.xlsx")
    P = at.Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=testdir / "databook_sir_dynamic.xlsx", make_default_parset=True, do_run=True)

    # # Do a scenario to get a second set of results
    par_results = P.results[-1]

    scvalues = dict()
    scen_par = "infdeath"
    scen_pop = "adults"
    scvalues[scen_par] = dict()
    scvalues[scen_par][scen_pop] = dict()
    scvalues[scen_par][scen_pop]["y"] = [0.2, 0.2]
    scvalues[scen_par][scen_pop]["t"] = [2014.0, 2050.0]
    scen = P.make_scenario(which="parameter", name="Increased mortality", scenario_values=scvalues)
    scen_results = scen.run(P, P.parsets["default"])
    par_results.name = "Baseline"
    scen_results.name = "Scenario"

    # Single cascades with data
    at.plot_cascade(par_results, cascade="main", pops="adults", year=2017, data=P.data)
    at.plot_cascade(scen_results, cascade="main", pops="adults", year=2017, data=P.data)

    # Single cascades without data
    at.plot_cascade(par_results, cascade="main", pops="adults", year=2025, data=P.data)
    at.plot_cascade(scen_results, cascade="main", pops="adults", year=2025, data=P.data)

    at.plot_multi_cascade([par_results, scen_results], cascade="main", pops="adults", year=[2017, 2025], data=P.data)

    d = at.PlotData(par_results, outputs=["sus", "inf", "rec", "dead"])
    at.plot_series(d, plot_type="stacked")

    d = at.PlotData(scen_results, outputs=["sus", "inf", "rec", "dead"])
    at.plot_series(d, plot_type="stacked")

    # Single cascade series
    at.plot_single_cascade_series(par_results, cascade="main", pops="adults", data=P.data)


if __name__ == "__main__":
    test_cascade_validate()
    test_cascade_basic_tb()
    test_cascade_basic_udt()
    test_cascade_scen_tb()
    test_cascade_scen_udt()
    test_cascade_dynamic()
    test_cascade_sir()



import atomica as at
import numpy as np
import pylab as pl
import matplotlib.pyplot as plt
from atomica.optimization import optimize

# np.seterr('raise')

F = at.ProjectFramework(at.LIBRARY_PATH / "malaria_framework.xlsx")
P = at.Project(framework=F, sim_dt=1.0 / 365.0)

# P.create_databook(databook_path=at.LIBRARY_PATH'malaria_databook.xlsx', num_pops=4, num_transfers=1, data_start=2010.)
P.load_databook(at.LIBRARY_PATH / "malaria_databook.xlsx", do_run=False)

# P.make_progbook(at.LIBRARY_PATH'malaria_progbook.xlsx', progs=17)
P.load_progbook(at.LIBRARY_PATH / "malaria_progbook.xlsx" "")

instructions = at.ProgramInstructions(2018)

res = P.run_sim("default", "default", instructions)
# P.save('malaria')
P.plot(res)

# Model is still in flux
# d = at.PlotData(res,'dalys',accumulate='integrate')
# at.plot_series(d,plot_type='stacked',axis='pops')
#
# # Number of treatments
# d = at.PlotData(res,'treated','gp')
# at.plot_series(d)
#
# # Eligibility for treatments
# d = at.PlotData(res,['hinf','himmmls','hwanmls'],'gp')
# at.plot_series(d,plot_type='stacked')
#
# # Flow into treatment eligibility
# d = at.PlotData(res,[':hinf',':himmmls',':hwanmls'],'gp')
# at.plot_series(d,plot_type='stacked')
#
# # Treated actual flow
# d = at.PlotData(res,'treated:flow','gp')
# at.plot_series(d)



import atomica as at
from atomica import ProjectFramework, Project
import os

testdir = at.parent_dir()
tmpdir = testdir / "temp"

F = ProjectFramework(testdir / "framework_sir_dynamic.xlsx")
P = Project(name="test", framework=F, do_run=False)
P.load_databook(databook_path=testdir / "databook_sir_dynamic.xlsx", make_default_parset=True, do_run=True)
result = P.results[0]

# # Do a scenario to get a second set of results
par_results = P.results[-1]

scvalues = dict()
scen_par = "infdeath"
scen_pop = "adults"
scvalues[scen_par] = dict()
scvalues[scen_par][scen_pop] = dict()
scvalues[scen_par][scen_pop]["y"] = [0.2, 0.2]
scvalues[scen_par][scen_pop]["t"] = [2014.0, 2050.0]
scen = P.make_scenario(which="parameter", name="Increased mortality", instructions=scvalues)
scen_results = scen.run(P, P.parsets["default"])
par_results.name = "Baseline"
scen_results.name = "Scenario"

d = at.PlotData([par_results, scen_results], outputs=["sus", "inf", "rec"], t_bins=10)
at.plot_bars(d, stack_outputs=[["sus", "inf"]], orientation="horizontal")



import os
import numpy as np
import atomica as at

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_combined_creation():

    F = at.ProjectFramework(at.LIBRARY_PATH / "combined_framework.xlsx")
    pops = {
        "SIR_0-4": {"label": "SIR 0-4", "type": "sir"},
        "SIR_5-14": {"label": "SIR 5-14", "type": "sir"},
        "SIR_15-64": {"label": "SIR 15-64", "type": "sir"},
        "UDT_0-14": {"label": "UDT 0-14", "type": "udt"},
        "UDT_15-64": {"label": "UDT 15-64", "type": "udt"},
    }

    transfers = {
        "aging_sir": {"label": "Aging SIR", "type": "sir"},
        "aging_udt": {"label": "Aging UDT", "type": "udt"},
    }

    D = at.ProjectData.new(framework=F, tvec=[2016, 2017, 2018], pops=pops, transfers=transfers)
    D.save(tmpdir / "combined_test.xlsx")
    D2 = at.ProjectData.from_spreadsheet(tmpdir / "combined_test.xlsx", framework=F)
    D2.add_pop("UDT_65+", "UDT 65+", "udt")
    D2.add_transfer("incarceration", "Incarceration", "sir")
    D2.save(tmpdir / "combined_test2.xlsx")

    D3 = at.ProjectData.from_spreadsheet(tmpdir / "combined_test.xlsx", framework=F)  # Load in the original one because it has no missing values
    D3.validate(F)

    P = at.Project(framework=F, databook=tmpdir / "combined_test.xlsx")

    ps = at.ProgramSet.new("default", tvec=np.arange(2015, 2019), progs=9, project=P)
    ps.save(tmpdir / "combined_progbook_test.xlsx")

    ps2 = at.ProgramSet.from_spreadsheet(at.LIBRARY_PATH / "combined_progbook.xlsx", project=P)
    ps2.add_program("test", "test")
    ps2.save(tmpdir / "combined_added.xlsx")


def test_combined_values():
    P1 = at.demo("sir", do_run=False)
    P1.settings.update_time_vector(start=2016, end=2021)
    R1 = P1.run_sim("default")
    P2 = at.demo("udt", do_run=False)
    P2.settings.update_time_vector(start=2016, end=2021)
    R2 = P2.run_sim("default")
    P3 = at.demo("combined", do_run=False)
    P3.settings.update_time_vector(start=2016, end=2021)
    R3 = P3.run_sim("default")

    # Check SIR values
    assert np.allclose(R1.model.pops[0].get_comp("sus").vals, R3.model.pops[0].get_comp("sus").vals, equal_nan=True)
    assert np.allclose(R1.model.pops[0].get_par("foi").vals, R3.model.pops[0].get_par("foi").vals, equal_nan=True)

    # Check UDT values
    assert np.allclose(R2.model.pops[0].get_comp("undx").vals, R3.model.pops[3].get_comp("undx").vals, equal_nan=True)

    # Check interaction values
    # Work through the interaction calculation between SIR and UDT
    interactions = P3.parsets[0].interactions[1]
    x = {}
    for to_pop in ["UDT_0-14", "UDT_15-64"]:
        x[to_pop] = np.zeros(R3.t.shape)
        for from_pop in ["SIR_0-4", "SIR_5-14", "SIR_15-64"]:
            x[to_pop] += R3.get_variable("foi", from_pop)[0].vals * interactions[from_pop].ts[to_pop].interpolate(R3.t)
    assert np.allclose(R3.get_variable("sum_foi", "UDT_0-14")[0].vals, x["UDT_0-14"], equal_nan=True)
    assert np.allclose(R3.get_variable("sum_foi", "UDT_15-64")[0].vals, x["UDT_15-64"], equal_nan=True)


def test_combined_cascades():
    P = at.demo("combined")
    at.get_cascade_vals(P.results[0], "sir_cascade")[0]
    udt_cascade = at.get_cascade_vals(P.results[0], "udt_cascade")[0]

    # Check aggregation is correctly applied
    udt_1pop_cascade = at.get_cascade_vals(P.results[0], "udt_cascade", pops="UDT_0-14")[0]
    udt_explicit_cascade = at.get_cascade_vals(P.results[0], "udt_cascade", pops=["UDT_0-14", "UDT_15-64"])[0]

    assert np.allclose(udt_cascade[0], 2 * udt_1pop_cascade[0], equal_nan=True)
    assert np.allclose(udt_cascade[0], udt_explicit_cascade[0], equal_nan=True)


def test_combined_plots():
    P = at.demo("combined")
    P.results[0].plot()


def test_combined_order():
    testdir = at.parent_dir()  # Must be relative to current file to work with tox
    P = at.Project(framework=testdir / "test_order_framework.xlsx", databook=testdir / "test_order_databook.xlsx")
    res = P.results[0]

    # This framework has interdependencies that mean the parameters must be resolved in exact framework order
    assert np.all(np.isfinite(res.get_variable("testpar")[0].vals))


if __name__ == "__main__":
    test_combined_creation()
    test_combined_values()
    test_combined_cascades()
    test_combined_plots()
    test_combined_order()



"""
Atomica root module

The Atomica module is the entry point for running simulations and performing analysis
using Atomica.

"""

# License:
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Lesser General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Display version information using logging

import sys
import logging

logger = logging.getLogger("atomica")

if not logger.handlers:
    # Only add handlers if they don't already exist in the module-level logger
    # This means that it's possible for the user to completely customize *a* logger called 'atomica'
    # prior to importing Atomica, and the user's custom logger won't be overwritten as long as it has
    # at least one handler already added. The use case was originally to suppress messages on import, but since
    # importing is silent now, it doesn't matter so much.
    debug_handler = logging.StreamHandler(sys.stdout)  # info_handler will handle all messages below WARNING sending them to STDOUT
    info_handler = logging.StreamHandler(sys.stdout)  # info_handler will handle all messages below WARNING sending them to STDOUT
    warning_handler = logging.StreamHandler(sys.stderr)  # warning_handler will send all messages at or above WARNING to STDERR

    debug_handler.setLevel(0)  # Handle all lower levels - the output should be filtered further by setting the logger level, not the handler level
    info_handler.setLevel(logging.INFO)  # Handle all lower levels - the output should be filtered further by setting the logger level, not the handler level
    warning_handler.setLevel(logging.WARNING)

    debug_handler.addFilter(type("ThresholdFilter", (object,), {"filter": lambda x, logRecord: logRecord.levelno < logging.INFO})())  # Display anything INFO or higher
    info_handler.addFilter(type("ThresholdFilter", (object,), {"filter": lambda x, logRecord: logRecord.levelno < logging.WARNING})())  # Don't display WARNING or higher

    warning_formatter = logging.Formatter("%(levelname)s {%(filename)s:%(lineno)d} - %(message)s")
    warning_handler.setFormatter(warning_formatter)

    debug_formatter = logging.Formatter("%(levelname)s {%(filename)s:%(lineno)d} - %(message)s")
    debug_handler.setFormatter(debug_formatter)

    logger.addHandler(debug_handler)
    logger.addHandler(info_handler)
    logger.addHandler(warning_handler)
    logger.setLevel("INFO")  # Set the overall log level


# Get version information
from .version import version as __version__, versiondate as __versiondate__
import sciris as sc
__gitinfo__ = sc.gitinfo(__file__)
version.gitinfo = __gitinfo__ # Add back (so version.py does not require imports)


# Check scipy version
import scipy
if sc.compareversions(scipy.__version__, "1.2.1") < 0:
    raise Exception(f"Atomica requires Scipy 1.2.1 or later - installed version is {scipy.__version__}")

# Increase Framework performance by not calling garbage collection all the time
import pandas as pd

pd.set_option("mode.chained_assignment", None)

# Suppress openpyxl deprecation warnings
import warnings

warnings.filterwarnings(action="ignore", category=UserWarning, module=".*openpyxl")


# The Atomica user interface -- import from submodules
from .calibration import *
from .cascade import *
from .data import *
from .demos import *
from .framework import *
from .function_parser import *
from .migration import migrations, register_migration
from .model import *
from .optimization import *
from .parameters import *
from .plotting import *
from .programs import *
from .project import *
from .reconciliation import *
from .results import *
from .scenarios import *
from .system import *
from .utils import *

import pathlib

rootdir = pathlib.Path(__file__).parent.parent



"""
Check whether automated model documentation template generation works
"""

import os
import atomica as at
import sciris as sc
import pytest

testdir = at.parent_dir()
tmpdir = testdir / "temp"
frameworks = list()
for f in at.LIBRARY_PATH.iterdir():
    if f.name.endswith("_framework.xlsx") and not f.name.startswith("~$"):
        frameworks.append(f)


@pytest.mark.parametrize("fname", frameworks)
def test_docgen(fname):
    F = at.ProjectFramework(fname)
    at.generate_framework_doc(F, fname=tmpdir / (fname.stem + "_doc.md"))


if __name__ == "__main__":
    for fname in frameworks:
        test_docgen(fname)



import atomica as at
import os
import sys


def test_migration():

    # The migration test tries to load a pickle saved in Python 3.7
    # This is known not to work in Python 3.6 because SafeUUID in the
    # UUID module was only introduced in Python 3.7. Therefore, we skip
    # this test in versions older than Python 3.7
    if sys.version_info[1] < 7:
        return

    at.logger.setLevel("DEBUG")

    testdir = at.parent_dir()
    tmpdir = testdir / "temp"

    P = at.Project.load(testdir / "migration_test_with_scenarios.prj")
    results = P.run_scenarios()
    P.progsets[0].save(tmpdir / "migration_test_progbook_save")  # Save original databook

    P = at.Project.load(testdir / "migration_test_with_result.prj")
    results = P.run_sim()
    at.plot_series(at.PlotData(results))

    P = at.Project.load(testdir / "migration_test_without_result.prj")
    results = P.run_sim()

    at.plot_series(at.PlotData(results))

    P.databook.save(tmpdir / "migration_test_databook_save")  # Save original databook
    P.data.save(tmpdir / "migration_test_data_save")  # Re-convert data to spreadsheet and save

    # Test databook operations
    P.data.add_pop("test_pop", "test_pop")  # Requires migration
    P.data.add_transfer("test_transfer", "test_transfer")
    P.data.add_interaction("test_interaction", "test_interaction")


if __name__ == "__main__":
    test_migration()



# -*- coding: utf-8 -*-
"""
Miscellaneous utility functions for Excel files

This module implements utility functions for Excel functionality that is common
to different kinds of spreadsheets used in Atomica (e.g. Databooks and Program Books).
For example, Excel formatting, and time-varying data entry tables, are implemented here.

"""

from xlsxwriter.utility import xl_rowcol_to_cell as xlrc
import sciris as sc
import io
import numpy as np
from .system import FrameworkSettings as FS
import pandas as pd
from .utils import format_duration, datetime_to_year
import xlsxwriter
from typing import Tuple
from openpyxl.utils import get_column_letter

__all__ = ["standard_formats", "apply_widths", "update_widths", "transfer_comments", "copy_sheet", "read_tables", "read_dataframes", "TimeDependentConnections", "TimeDependentValuesEntry", "cell_get_string", "cell_get_number", "validate_category"]

# Suppress known warning in Openpyxl
# Warnings are:
# - C:\ProgramData\Miniconda3\envs\atomica37\lib\site-packages\openpyxl\worksheet\_reader.py:300: UserWarning: Conditional Formatting extension is not supported and will be removed
#   warn(msg)
# - C:\ProgramData\Miniconda3\envs\atomica37\lib\site-packages\openpyxl\worksheet\_reader.py:300: UserWarning: Data Validation extension is not supported and will be removed
#   warn(msg)
# This means that conditional formatting and data valuation rules aren't being loaded, but since `data_only=True` these don't matter and can be safely ignored
import warnings

warnings.filterwarnings(action="ignore", category=UserWarning, module="openpyxl.worksheet", lineno=300)


def standard_formats(workbook: xlsxwriter.Workbook) -> dict:
    """
    Add standard formatting to a workbook

    Returns a dictionary of format objects for use when writing the workbook

    :param workbook: an xlsxwriter workbook
    :return: A dictionary of ``xlsxwriter.workbook.Format`` instances
    """
    atomica_blue = "#98E0FA"
    optional_orange = "#FFA500"
    BG_COLOR = atomica_blue
    OPT_COLOR = optional_orange
    BORDER_COLOR = "white"

    formats = {}

    # Locked formats
    formats["bold"] = workbook.add_format({"bold": 1})
    formats["center"] = workbook.add_format({"align": "center"})
    formats["center_bold"] = workbook.add_format({"bold": 1, "align": "center"})
    formats["rc_title"] = {}
    formats["rc_title"]["right"] = {}
    formats["rc_title"]["right"]["T"] = workbook.add_format({"bold": 1, "align": "right", "text_wrap": True})
    formats["rc_title"]["right"]["F"] = workbook.add_format({"bold": 1, "align": "right", "text_wrap": False})
    formats["rc_title"]["left"] = {}
    formats["rc_title"]["left"]["T"] = workbook.add_format({"bold": 1, "align": "left", "text_wrap": True})
    formats["rc_title"]["left"]["F"] = workbook.add_format({"bold": 1, "align": "left", "text_wrap": False})

    # Unlocked formats
    formats["unlocked"] = workbook.add_format({"locked": 0, "bg_color": BG_COLOR, "border": 1, "border_color": BORDER_COLOR})
    formats["center_unlocked"] = workbook.add_format({"align": "center", "locked": 0, "bg_color": BG_COLOR, "border": 1, "border_color": BORDER_COLOR})
    formats["general"] = workbook.add_format({"locked": 0, "num_format": 0x00, "bg_color": BG_COLOR, "border": 1, "border_color": BORDER_COLOR})

    # Conditional formats
    formats["unlocked_boolean_true"] = workbook.add_format({"bg_color": OPT_COLOR})
    formats["unlocked_boolean_false"] = workbook.add_format({"bg_color": BG_COLOR})
    formats["not_required"] = workbook.add_format({"bg_color": "#EEEEEE", "border": 1, "border_color": "#CCCCCC"})
    formats["white_bg"] = workbook.add_format({"bg_color": "#FFFFFF", "border": 1, "border_color": "#CCCCCC"})
    formats["ignored"] = workbook.add_format({"pattern": 14})  # Hatched with diagonal lines - this represents a cell whose value will not be used in the model run (e.g., an assumption that also has time-specific points)
    formats["warning"] = workbook.add_format({"bg_color": "#FF0000"})
    formats["ignored_warning"] = workbook.add_format({"pattern": 14, "bg_color": "#FF0000"})  # hatched, with red background
    formats["ignored_not_required"] = workbook.add_format({"pattern": 14, "bg_color": "#EEEEEE", "border": 1, "border_color": "#CCCCCC"})  # hatched, with grey background

    return formats


def apply_widths(worksheet: xlsxwriter.workbook.Worksheet, width_dict: dict) -> None:
    """
    Set column widths

    Given a dictionary of precomputed character counts (typically the maximum number of characters in
    any cell written to each column), set the width of the column in Excel accordingly. The Excel
    column width is slightly larger than the character count.

    :param worksheet: A Worksheet instance
    :param width_dict: A dictionary of character widths to use for each column (by index)
    """

    for idx, width in width_dict.items():
        worksheet.set_column(idx, idx, width * 1.1 + 1)


def update_widths(width_dict: dict, column_index: int, contents: str) -> None:
    """
    Keep track of required width for a column

    ``width_dict`` is a dict that is keyed by column index e.g. 0,1,2
    and the value is the length of the longest contents seen for that column

    :param width_dict: Storage dictionary
    :param column_index: Index of the column value has been inserted in
    :param contents: Content, length of which is used to set width

    """

    if width_dict is None or contents is None or not sc.isstring(contents):
        return

    if len(contents) == 0:
        return

    if column_index not in width_dict:
        width_dict[column_index] = len(contents)
    else:
        width_dict[column_index] = max(width_dict[column_index], len(contents))


def transfer_comments(target: sc.Spreadsheet, comment_source: sc.Spreadsheet) -> None:
    """
    Copy comments between spreadsheets

    This function copies comments from one spreadsheet to another. Under the hood,
    a new spreadsheet is created with values from the ``target`` Spreadsheet
    and cell-wise formatting from the ``comment_source`` Spreadsheet. If a cell exists in
    this spreadsheet and not in the source, it will be retained as-is. If more cells exist in
    the ``comment_source`` than in this spreadsheet, those cells will be dropped. If a sheet exists in
    the ``comment_source`` and not in the current workbook, it will be added

    :param target: The target spreadsheet to write comments into
    :param comment_source: The source spreadsheet containing comments

    """

    import openpyxl
    from openpyxl.comments import Comment

    assert isinstance(target, sc.Spreadsheet)
    assert isinstance(comment_source, sc.Spreadsheet)

    this_workbook = openpyxl.load_workbook(target.tofile(), data_only=False)  # This is the value source workbook
    old_workbook = openpyxl.load_workbook(comment_source.tofile(), data_only=False)  # A openpyxl workbook for the old content

    for sheet in this_workbook.worksheets:

        # If this sheet isn't in the old workbook, do nothing
        if sheet.title not in old_workbook.sheetnames:
            continue

        # Transfer comments
        for row in old_workbook[sheet.title].rows:
            for cell in row:
                if cell.comment:
                    sheet[cell.coordinate].comment = Comment(cell.comment.text, "")

    # Save the modified spreadsheet to a new buffer
    f = io.BytesIO()
    this_workbook.save(f)
    f.flush()
    f.seek(0)
    target.load(f)


def copy_sheet(source: str, sheet_name: str, workbook: xlsxwriter.Workbook) -> None:
    """
    Copy a sheet into a Workbook

    This function allows values to be copied from a file into a Workbook. The
    main use case is to support custom sheets in databooks that are not otherwise parsed
    but which might need to be retained. In particular, the ``ProjectData`` class does
    not parse ignored sheets at all, because no guarantees can be made about the quantity and
    type of the content, and whether there are formulas etc. that would be lost. In some cases
    though, it may be necessary to edit the databook and preserve specific sheets. In general,
    this can be done by using the ``to_workbook()`` method and then manually performing whatever
    operations are necessary to preserve the content on the extra sheets. However, when the
    extra sheet content is extremely simple e.g. just a table of values, then this helper
    function can be used to facilitate copying the content.

    Warning - note that Excel functions, formatting, and comments will NOT be preserved.

    :param source: File name of the spreadsheet to read the source sheet from or an ``sc.Spreadsheet`` instance
    :param sheet_name: Name of the sheet to write to
    :param workbook: A Workbook instance to add the sheet to
    :return: None - the sheet will be added to the Workbook in-place
    """

    import openpyxl

    if sc.isstring(source):
        source = sc.Spreadsheet(source)

    src_workbook = openpyxl.load_workbook(source.tofile(), read_only=True, data_only=True)  # Load in read-only mode for performance, since we don't parse comments etc.
    src_worksheet = src_workbook[sheet_name]
    dst_worksheet = workbook.add_worksheet(sheet_name)

    for i, row in enumerate(src_worksheet.rows):
        for j, cell in enumerate(row):
            dst_worksheet.write(i, j, cell.value)

    src_workbook.close()


def read_tables(worksheet) -> Tuple[list, list]:
    """
    Read tables from sheet

    :param worksheet: An openpyxl worksheet
    :return: A tuple containing - A list of tables (which is a list of rows, terminated by an empty row in the original spreadsheet),
             and a list of start row indices for each table read in
    """
    # This function takes in a openpyxl worksheet, and returns tables
    # A table consists of a block of rows with any #ignore rows skipped
    # This function will start at the top of the worksheet, read rows into a buffer
    # until it gets to the first entirely empty row
    # And then returns the contents of that buffer as a table. So a table is a list of openpyxl rows
    # This function continues until it has exhausted all of the rows in the sheet

    buffer = []
    tables = []
    start_rows = []
    start = None

    for i, row in enumerate(worksheet.rows):

        # Determine whether to skip the row, add it to the buffer, or flush buffer into a table
        flush = False
        for j, cell in enumerate(row):
            if cell.value:
                if cell.data_type == "s" and cell.value.startswith("#ignore"):
                    if j == 0:
                        break  # If #ignore is encountered in the first column, skip the row and continue parsing the table
                    else:
                        flush = True
                        break  # If #ignore is encoutered after preceding blank cells
                else:
                    # Read the row into the buffer and continue parsing the table
                    if not buffer:
                        start = i + 1  # Excel rows are indexed starting at 1
                    buffer.append(row)
                    break  # If the cell has a value in it, continue parsing the table
        else:
            if buffer:
                flush = True

        if flush:
            tables.append(buffer)  # Only append the buffer if it is not empty
            start_rows.append(start)
            buffer = []

    # After the last row, if the buffer has some un-flushed contents, then include it in the last table
    if buffer:
        tables.append(buffer)
        start_rows.append(start)

    return tables, start_rows


def read_dataframes(worksheet, merge=False) -> list:
    """
    Read dataframes from sheet

    This function operates similarly to ``read_tables`` except it returns Dataframes instead of
    cells. This enables the dataframes to be constructed more quickly, at the expense of being
    able to track the cell references and row numbers. These are shown for databooks (via ``read_tables``)
    but not for frameworks (which go via ``read_dataframes``)


    :param worksheet: An openpyxl worksheet
    :param merge: If False (default) then blank rows will be used to split the dataframes. If True, only one
                  DataFrame will be returned
    :return: A list of DataFrames

    """

    if worksheet.max_row is None or worksheet.max_column is None:
        worksheet.calculate_dimension(force=True)

    content = np.empty((worksheet.max_row, worksheet.max_column), dtype="object")
    ignore = np.zeros((worksheet.max_row), dtype=bool)
    empty = np.zeros((worksheet.max_row), dtype=bool)  # True for index where a new table begins

    for i, row in enumerate(worksheet.rows):

        any_values = False  # Set True if this row contains any values
        for j, cell in enumerate(row):
            v = cell.value
            if cell.data_type == "s":
                if not any_values and v.startswith("#ignore"):
                    # If we encounter a #ignore and it's the first content in the row
                    if j == 0:
                        # If it's the first cell, ignore the row (i.e., do NOT treat it as a blank row)
                        ignore[i] = True  # Ignore the row
                    break
                elif v.startswith("#ignore"):
                    # Skip the rest of the row
                    content[i, j:] = None
                    break
                else:
                    v = v.strip()
                    any_values = any_values or bool(v)  # If it's a string type, call strip() before checking truthiness
            elif v is not None:
                any_values = True
            content[i, j] = v

        if not any_values:
            empty[i] = True

    tables = []
    if merge:
        ignore[empty] = True
        if all(ignore):
            return []
        tables.append(content[~ignore, :])
    else:
        # A change from False to True means that we need to start a new table
        # A True followed by a True doesn't start a new table but instead gets ignored
        content = content[~ignore, :]
        empty = empty[~ignore]

        # If there is no content at all, return immediately
        if all(empty):
            return []

        idx = []
        start = None

        for i in range(len(empty)):

            # Check for the start of a table
            if not empty[i] and start is None:
                start = i

            # Check for the end of a table
            if empty[i] and start is not None:
                # If we were reading a table and have reached an empty row
                idx.append((start, i))
                start = None
            elif i + 1 == len(empty) and start is not None:
                # If we were reading a table and have reached the end of the data
                idx.append((start, i + 1))
                start = None

        tables = []
        for start, stop in idx:
            tables.append(content[start:stop].copy())  # Use .copy() so that we don't retain any references to the original full array outside this function

    dfs = []
    for table in tables:
        df = pd.DataFrame(table)
        df.dropna(axis=1, how="all", inplace=True)
        df.columns = df.iloc[0]
        df = df[1:]
        dfs.append(df)
    return dfs


class TimeDependentConnections:
    """
    Structure for reading/writing interactions

    A :class:`TimeDependentConnections` object is suitable when there are time dependent interactions between two quantities
    This class is used for both transfers and interactions. The content that it writes consists of

    - A connection matrix table that has Y/N selection of which interactions are present between two things
    - A set of pairwise connections specifying to, from, units, assumption, and time

    Interactions can have a diagonal, whereas transfers cannot (e.g. a population can infect itself but cannot transfer to itself).

    In Excel, a :class:`TimeDependentConnections` maps to three tables

    1. A table to enter the code name and full name
    2. An interactions matrix with Y/N indicating whether an interaction between two populations exists
    3. A set of rows for entering time varying data for each pair of populations

    :param code_name: the code name of this quantity e.g. 'aging'
    :param full_name: the full name of this quantity e.g. 'Aging'
    :param tvec: time values for the time-dependent rows
    :param pops: list of strings to use as the rows and columns - these are typically lists of population code names
    :param type: 'transfer' or 'interaction'. A transfer cannot have diagonal entries, and can have Number or Probability formats. An Interaction can have
                 diagonal entries and only has N.A. formats
    :param ts: Optionally specify a dict containing all of the non-empty TimeSeries objects used. The format is ``{(from_pop, to_pop):TimeSeries}``.
               An interaction can only be Y/N for clarity, if it is Y then a row is displayed for the TimeSeries. Actually, the Y/N can be
               decided in the first instance based on the provided TimeSeries i.e. if a TimeSeries is provided for an interaction, then the
               interaction must have been marked with Y
    :param pop_type: Specify pop_type, which is used by :meth:`ProjectData.add_pop` to determine which TDCs to add new populations to

    """

    def __init__(self, code_name: str, full_name: str, tvec: np.array, from_pops: list, to_pops: list, interpop_type: str, ts: dict = None, from_pop_type: str = None, to_pop_type: str = None):
        self.code_name = code_name
        self.full_name = full_name
        self.type = interpop_type
        self.from_pop_type = from_pop_type
        self.from_pops = from_pops
        self.to_pop_type = to_pop_type
        self.to_pops = to_pops
        self.tvec = tvec
        self.ts = ts if ts is not None else sc.odict()

        self.attributes = {}  #: Attributes associated with the table
        self.ts_attributes = {}  #: Attributes associated with each TimeSeries row. This is a dict-of-dicts with `ts_attributes[attribute]={(from_pop, to_pop):value}` e.g., `ts_attributes["Provenance"] = {('0-4', '5-14'): 'value'}`
        self.ts_attributes["Provenance"] = {}  # Include provenance attribute by default

        self.assumption_heading = "Constant"  #: Heading to use for assumption column
        self.write_units = None  #: Write a column for units (if None, units will be written if any of the TimeSeries have units)
        self.write_uncertainty = None  #: Write a column for units (if None, units will be written if any of the TimeSeries have uncertainty)
        self.write_assumption = None  #: Write a column for units (if None, units will be written if any of the TimeSeries have an assumption)

        if self.type == "transfer":
            self.enable_diagonal = False
            self.allowed_units = []
            self.allowed_units.append("%s (%s)" % (FS.QUANTITY_TYPE_NUMBER.title(), format_duration(1, pluralize=True)))
            self.allowed_units.append("%s (per %s)" % (FS.QUANTITY_TYPE_RATE.title(), format_duration(1, pluralize=False)))
            self.allowed_units.append("%s (%s)" % (FS.QUANTITY_TYPE_DURATION.title(), format_duration(1, pluralize=True)))
        elif self.type == "interaction":
            self.enable_diagonal = True
            self.allowed_units = [FS.DEFAULT_SYMBOL_INAPPLICABLE]
        else:
            raise Exception('Unknown TimeDependentConnections type - must be "transfer" or "interaction"')

    def __repr__(self):
        return '<TDC %s "%s">' % (self.type.title(), self.code_name)

    @classmethod
    def from_tables(cls, tables: list, interaction_type):
        """
        Instantiate based on list of tables

        This method instantiates and initializes a new :class:`TimeDependentConnections` object from
        tables that have been read in using :func:`read_tables`. Note that the parent object
        such as :class:`ProjectData` is responsible for extracting the tables and passing them
        to this function. For instance, the transfers sheet might contain more than one set of
        tables, so it is the calling function's responsibility to split those tables up into
        the groups of three expected by this method.

        :param tables: A list of tables. A table here is a list of rows, and a row is a list of cells.
        :param interaction_type: A string identifying the interaction type - either 'transfer' or 'interaction'
        :return: A new :class:`TimeDependentConnections` instance

        """

        from .utils import TimeSeries  # Import here to avoid circular reference

        assert interaction_type in {"transfer", "interaction"}, "Unknown interaction type"

        # Read the TDC definition table (including attributes)
        code_name = None
        full_name = None
        from_pop_type = None
        to_pop_type = None

        # Start by reading the TDC header row specifying the name and pop types
        # This table also contains only a single row. Any subsequent rows will automatically be ignored
        attributes = {}
        for header_cell, value_cell in zip(tables[0][0], tables[0][1]):

            header = cell_get_string(header_cell, allow_empty=True)
            if header is None:
                continue
            elif header.startswith("#ignore"):
                break

            lowered_header = header.lower()
            if lowered_header == "abbreviation":
                code_name = cell_get_string(value_cell)
            elif lowered_header == "full name":
                full_name = cell_get_string(value_cell)
            elif lowered_header == "from population type":
                from_pop_type = cell_get_string(value_cell, True)
            elif lowered_header == "to population type":
                to_pop_type = cell_get_string(value_cell, True)
            else:
                attributes[header] = value_cell.value

        if interaction_type == "transfer":
            assert from_pop_type == to_pop_type, "Transfers can only occur between populations of the same type"

        if code_name is None:
            raise Exception("Code name/abbreviation missing")

        if full_name is None:
            raise Exception("Full name missing")

        # Read the pops from the Y/N table. The Y/N content of the table depends on the timeseries objects that
        # are present. That is, if the Y/N matrix contains a Y then a TimeSeries must be read in, and vice versa.
        # Therefore, we don't actually parse the matrix, and instead just read in all the TimeSeries instances
        # that are defined and infer the matrix that way. However, we do still need to parse the to_pops and from_pops
        # These both get parsed because for interactions across pop types they could be different. Blank cells in either
        # the row or the column signal the end of parsing those population lists

        # Read the to_pops from the first row of the table, until a blank cell is encountered
        to_pops = []
        for cell in tables[1][0][1:]:
            if cell.value is None or (cell.data_type == "s" and cell.value.startswith("#ignore")):
                break
            else:
                to_pops.append(cell.value)

        # Read the from_pops from the first column, until a blank cell is encountered. Note that a #ignore
        # can never be present in the first column, because this would cause the row to be skipped prior to this point
        from_pops = []
        for row in tables[1][1:]:
            if row[0].value is None:
                break
            else:
                from_pops.append(row[0].value)

        # Instantiate the TDC object based on the metadata from the first two tables
        tdc = cls(code_name, full_name, None, from_pops=from_pops, to_pops=to_pops, interpop_type=interaction_type, from_pop_type=from_pop_type, to_pop_type=to_pop_type)
        tdc.attributes = attributes

        known_headings = {"from population", "to population", "units", "uncertainty", "constant", "assumption"}
        headings, times, tdc.tvec = _parse_ts_header(tables[2][0], known_headings, skip_first=False)

        # Validate and process headings
        if not times and "constant" not in headings:
            raise Exception("Could not find an assumption or time-specific value - all tables must contain at least one of these values")
        tdc.write_units = True if "units" in headings else None
        tdc.write_uncertainty = True if "uncertainty" in headings else None
        tdc.write_assumption = True if "constant" in headings else None
        if "assumption" in headings:
            tdc.write_assumption = True
            tdc.assumption_heading = "Assumption"
        for heading in headings:
            if heading not in known_headings:
                tdc.ts_attributes[heading] = {}

        tdc.ts = sc.odict()
        for row in tables[2][1:]:
            if row[0].value != "...":
                assert row[0].value in from_pops, 'Population "%s" not found - should be contained in %s' % (row[0].value, from_pops)
                assert row[2].value in to_pops, 'Population "%s" not found - should be contained in %s' % (row[2].value, to_pops)
                vals = [x.value for x in row]
                from_pop = vals[0]
                to_pop = vals[2]

                if "units" in headings:
                    units = cell_get_string(row[headings["units"]], allow_empty=True)
                    if units is not None and units.lower().strip() in FS.STANDARD_UNITS:
                        units = units.lower().strip()  # Only lower and strip units if they are standard units
                else:
                    units = None
                ts = TimeSeries(units=units)

                if "uncertainty" in headings:
                    ts.sigma = cell_get_number(row[headings["uncertainty"]])
                else:
                    ts.sigma = None

                if "constant" in headings:
                    ts.assumption = cell_get_number(row[headings["constant"]])
                elif "assumption" in headings:
                    ts.assumption = cell_get_number(row[headings["assumption"]])
                else:
                    ts.assumption = None

                for attribute in tdc.ts_attributes:
                    if attribute in headings:
                        tdc.ts_attributes[attribute][(from_pop, to_pop)] = row[headings[attribute]].value

                for t, idx in times.items():
                    ts.insert(t, cell_get_number(row[idx]))  # If cell_get_number returns None, this gets handled accordingly by ts.insert()

                tdc.ts[(from_pop, to_pop)] = ts

        return tdc

    def write(self, worksheet, start_row, formats, references: dict = None, widths: dict = None) -> int:
        """
        Write to cells in a worksheet

        :param worksheet: An xlsxwriter worksheet instance
        :param start_row: The first row in which to write values
        :param formats: Format dict for the opened workbook - typically the return value of :func:`standard_formats` when the workbook was opened
        :param references: References dict containing cell references for strings in the current workbook
        :param widths: ``dict`` storing column widths
        :return: The row index for the next available row for writing in the spreadsheet

        """

        assert self.assumption_heading in {"Constant", "Assumption"}, "Unsupported assumption heading"
        write_units = self.write_units if self.write_units is not None else any((ts.units is not None for ts in self.ts.values()))
        write_uncertainty = self.write_uncertainty if self.write_uncertainty is not None else any((ts.sigma is not None for ts in self.ts.values()))
        write_assumption = self.write_assumption if self.write_assumption is not None else any((ts.assumption is not None for ts in self.ts.values()))

        if not references:
            references = {x: x for x in self.from_pops + self.to_pops}  # Default null mapping for populations

        # First, write the name entry table
        current_row = start_row

        column = 0
        worksheet.write(current_row, column, "Abbreviation", formats["center_bold"])
        update_widths(widths, column, "Abbreviation")
        worksheet.write(current_row + 1, column, self.code_name)
        update_widths(widths, column, self.code_name)

        column += 1
        worksheet.write(current_row, column, "Full Name", formats["center_bold"])
        update_widths(widths, column, "Full Name")
        worksheet.write(current_row + 1, column, self.full_name)
        update_widths(widths, column, self.full_name)

        column += 1
        worksheet.write(current_row, column, "From population type", formats["center_bold"])
        update_widths(widths, column, "From population type")
        worksheet.write(current_row + 1, column, self.from_pop_type)
        update_widths(widths, column, self.from_pop_type)

        column += 1
        worksheet.write(current_row, column, "To population type", formats["center_bold"])
        update_widths(widths, column, "To population type")
        worksheet.write(current_row + 1, column, self.to_pop_type)
        update_widths(widths, column, self.to_pop_type)

        for attribute, value in self.attributes.items():
            column += 1
            worksheet.write(current_row, column, attribute, formats["center_bold"])
            update_widths(widths, column, attribute)
            worksheet.write(current_row + 1, column, value)
            update_widths(widths, column, value)

        # Then, write the Y/N matrix
        current_row += 3  # Leave a blank row below the matrix
        # Note - table_references are local to this TimeDependentConnections instance
        # For example, there could be two transfers, and each of them could potentially transfer between 0-4 and 5-14
        # so the worksheet might contain two references from 0-4 to 5-14 but they would be for different transfers and thus
        # the time-dependent rows would depend on different boolean table cells
        current_row, table_references, values_written = self._write_pop_matrix(worksheet, current_row, formats, references, boolean_choice=True, widths=widths)

        # Finally, write the time dependent part
        headings = []
        headings.append("From population")
        headings.append("")  # --->
        headings.append("To population")
        offset = len(headings)

        attribute_index = {}
        for attribute in self.ts_attributes:
            attribute_index[attribute] = offset
            headings.append(attribute)
            offset += 1

        if write_units:
            headings.append("Units")
            units_index = offset  # Column to write the units in
            offset += 1

        if write_uncertainty:
            headings.append("Uncertainty")
            uncertainty_index = offset  # Column to write the units in
            offset += 1

        if write_assumption:
            headings.append(self.assumption_heading)
            headings.append("")
            constant_index = offset
            offset += 1
            if len(self.tvec):
                offset += 1  # Additional offset for the 'OR' column

        headings += [float(x) for x in self.tvec]
        for i, entry in enumerate(headings):
            worksheet.write(current_row, i, entry, formats["center_bold"])
            update_widths(widths, i, entry)

        # Now, we will write a wrapper that gates the content
        # If the gating cell is 'Y', then the content will be displayed, otherwise not
        def gate_content(content, gating_cell):
            if content.startswith("="):  # If this is itself a reference
                return '=IF(%s="Y",%s,"...")' % (gating_cell, content[1:])
            else:
                return '=IF(%s="Y","%s","...")' % (gating_cell, content)

        for from_idx in range(0, len(self.from_pops)):
            for to_idx in range(0, len(self.to_pops)):
                current_row += 1
                from_pop = self.from_pops[from_idx]
                to_pop = self.to_pops[to_idx]
                entry_tuple = (from_pop, to_pop)
                entry_cell = table_references[entry_tuple]

                # Write hyperlink
                if values_written[entry_cell] != FS.DEFAULT_SYMBOL_INAPPLICABLE:
                    worksheet.write_url(entry_cell, "internal:%s!%s" % (worksheet.name, xlrc(current_row, 1)), cell_format=formats["center_unlocked"], string=values_written[entry_cell])
                    worksheet.write_url(xlrc(current_row, 1), "internal:%s!%s" % (worksheet.name, entry_cell), cell_format=formats["center_unlocked"])

                if entry_tuple in self.ts:
                    ts = self.ts[entry_tuple]
                    format = formats["not_required"]
                else:
                    ts = None
                    format = formats["unlocked"]

                if ts:
                    worksheet.write_formula(current_row, 0, gate_content(references[from_pop], entry_cell), formats["center_bold"], value=from_pop)
                    update_widths(widths, 0, from_pop)
                    worksheet.write_formula(current_row, 1, gate_content("--->", entry_cell), formats["center"], value="--->")
                    worksheet.write_formula(current_row, 2, gate_content(references[to_pop], entry_cell), formats["center_bold"], value=to_pop)
                    update_widths(widths, 2, to_pop)

                    # Write the attributes
                    for attribute in self.ts_attributes:
                        if isinstance(self.ts_attributes[attribute], dict):
                            if entry_tuple in self.ts_attributes[attribute]:
                                val = self.ts_attributes[attribute][entry_tuple]
                            else:
                                val = None
                        else:
                            val = self.ts_attributes[attribute]

                        if val is not None:
                            worksheet.write(current_row, attribute_index[attribute], val)
                            update_widths(widths, attribute_index[attribute], val)

                    if self.write_units:
                        worksheet.write(current_row, units_index, ts.units, format)
                        update_widths(widths, units_index, ts.units)
                        if self.allowed_units:
                            worksheet.data_validation(xlrc(current_row, units_index), {"validate": "list", "source": [x for x in self.allowed_units]})

                    if self.write_uncertainty:
                        worksheet.write(current_row, uncertainty_index, ts.sigma, formats["not_required"])

                    if self.write_assumption:
                        worksheet.write(current_row, constant_index, ts.assumption, format)
                        if len(self.tvec):
                            worksheet.write_formula(current_row, constant_index + 1, gate_content("OR", entry_cell), formats["center"], value="OR")
                            update_widths(widths, constant_index + 1, "OR")

                else:
                    worksheet.write_formula(current_row, 0, gate_content(references[from_pop], entry_cell), formats["center_bold"], value="...")
                    worksheet.write_formula(current_row, 1, gate_content("--->", entry_cell), formats["center"], value="...")
                    worksheet.write_formula(current_row, 2, gate_content(references[to_pop], entry_cell), formats["center_bold"], value="...")

                    if self.write_units:
                        worksheet.write_blank(current_row, units_index, "", format)
                        if self.allowed_units:
                            worksheet.data_validation(xlrc(current_row, units_index), {"validate": "list", "source": [x for x in self.allowed_units]})

                    if self.write_uncertainty:
                        worksheet.write_blank(current_row, uncertainty_index, "", formats["not_required"])

                    if self.write_assumption:
                        worksheet.write_blank(current_row, constant_index, "", format)
                        if len(self.tvec):
                            worksheet.write_formula(current_row, constant_index + 1, gate_content("OR", entry_cell), formats["center"], value="...")
                            update_widths(widths, constant_index + 1, "...")

                content = [None] * len(self.tvec)

                if ts:
                    for t, v in zip(ts.t, ts.vals):
                        idx = np.where(self.tvec == t)[0][0]
                        content[idx] = v

                for idx, v in enumerate(content):
                    if v is None:
                        worksheet.write_blank(current_row, offset + idx, v, format)
                    else:
                        worksheet.write(current_row, offset + idx, v, format)
                    widths[offset + idx] = max(widths[offset + idx], 7) if offset + idx in widths else 7

                if not content:
                    idx = 0

                if self.write_assumption and len(self.tvec):
                    # Conditional formatting for the assumption, depending on whether time-values were entered
                    fcn_empty_times = 'COUNTIF(%s:%s,"<>" & "")>0' % (xlrc(current_row, offset), xlrc(current_row, offset + idx))
                    worksheet.conditional_format(xlrc(current_row, constant_index), {"type": "formula", "criteria": "=" + fcn_empty_times, "format": formats["ignored"]})
                    worksheet.conditional_format(xlrc(current_row, constant_index), {"type": "formula", "criteria": "=AND(%s,NOT(ISBLANK(%s)))" % (fcn_empty_times, xlrc(current_row, constant_index)), "format": formats["ignored_warning"]})

                # Conditional formatting for the row - it has a white background if the gating cell is 'N'
                worksheet.conditional_format("%s:%s" % (xlrc(current_row, 3), xlrc(current_row, offset + idx)), {"type": "formula", "criteria": '=%s<>"Y"' % (entry_cell), "format": formats["white_bg"]})

        current_row += 2

        return current_row

    def _write_pop_matrix(self, worksheet, start_row, formats, references: dict = None, boolean_choice=False, widths: dict = None):
        """
        Write a square matrix to Excel

        This function writes the Y/N matrix

        - Transfer matrix
        - Interactions matrix

        If ``self.enable_diagonal`` is ``False`` then the diagonal will be forced to be ``'N.A.'``. If an entry
        is specified for an entry on the diagonal and ``enable_diagonal=False``, an error will be thrown

        :param worksheet: An xlsxwriter worksheet instance
        :param start_row: The first row in which to write values
        :param formats: Format dict for the opened workbook - typically the return value of :func:`standard_formats` when the workbook was opened
        :param references: Optionally supply dict with references, used to link population names in Excel
        :param boolean_choice: If True, values will be coerced to Y/N and an Excel validation will be added
        :param widths: ``dict`` storing column widths
        :return: Tuple with ``(next_row, table_references, values_written)``. The references are used for hyperlinking to the Excel matrix

        """

        entries = self.ts

        if not references:
            references = {x: x for x in self.from_pops + self.to_pops}  # This is a null-mapping that takes say 'adults'->'adults' thus simplifying the workflow. Otherwise, it's assumed a reference exists for every node

        table_references = {}
        values_written = {}

        # Write the headers
        for i, node in enumerate(self.to_pops):
            worksheet.write_formula(start_row, i + 1, references[node], formats["center_bold"], value=node)
            update_widths(widths, i + 1, node)
        for i, node in enumerate(self.from_pops):
            worksheet.write_formula(start_row + i + 1, 0, references[node], formats["center_bold"], value=node)
            update_widths(widths, 0, node)

        # Prepare the content - first replace the dict with one keyed by index. This is because we cannot apply formatting
        # after writing content, so have to do the writing in a single pass over the entire matrix
        if boolean_choice:
            content = np.full((len(self.from_pops), len(self.to_pops)), "N", dtype=object)  # This will also coerce the value to string in preparation for writing
        else:
            content = np.full((len(self.from_pops), len(self.to_pops)), "", dtype=object)  # This will also coerce the value to string in preparation for writing

        for interaction, value in entries.items():
            from_pop, to_pop = interaction
            if not self.enable_diagonal and from_pop == to_pop:
                raise Exception("Trying to write a diagonal entry to a table that is not allowed to contain diagonal terms")  # This is because data loss will occur if the user adds entries on the diagonal, then writes the table, and then reads it back in
            from_idx = self.from_pops.index(from_pop)
            to_idx = self.to_pops.index(to_pop)
            if boolean_choice:
                value = "Y" if value else "N"
            content[from_idx, to_idx] = value

        # Write the content
        for from_idx in range(0, len(self.from_pops)):
            for to_idx in range(0, len(self.to_pops)):
                row = start_row + 1 + from_idx
                col = to_idx + 1
                if not self.enable_diagonal and self.to_pops[to_idx] == self.from_pops[from_idx]:  # Disable the diagonal if it's linking the same two quantities and that's desired
                    val = FS.DEFAULT_SYMBOL_INAPPLICABLE
                    worksheet.write(row, col, val, formats["center"])
                    worksheet.data_validation(xlrc(row, col), {"validate": "list", "source": ["N.A."]})
                else:
                    val = content[from_idx, to_idx]
                    worksheet.write(row, col, content[from_idx, to_idx], formats["center_unlocked"])
                    if boolean_choice:
                        worksheet.data_validation(xlrc(row, col), {"validate": "list", "source": ["Y", "N"]})
                        worksheet.conditional_format(xlrc(row, col), {"type": "cell", "criteria": "equal to", "value": '"Y"', "format": formats["unlocked_boolean_true"]})
                        worksheet.conditional_format(xlrc(row, col), {"type": "cell", "criteria": "equal to", "value": '"N"', "format": formats["unlocked_boolean_false"]})
                table_references[(self.from_pops[from_idx], self.to_pops[to_idx])] = xlrc(row, col, True, True)  # Store reference to this interaction
                values_written[table_references[(self.from_pops[from_idx], self.to_pops[to_idx])]] = val

        next_row = start_row + 1 + len(self.from_pops) + 1
        return next_row, table_references, values_written


class TimeDependentValuesEntry:
    """Table for time-dependent data entry

    This class is Databooks and Program books to enter potentially time-varying data.
    Conceptually, it maps a set of TimeSeries object to a single name and table in the
    spreadsheet. For example, a Characteristic might contain a TimeSeries for each population,
    and the resulting TimeDependentValuesEntry (TDVE) table would have a `name` matching the
    population, and TimeSeries for each population.

    The TDVE class optionally allows the specification of units, assumptions, and uncertainty,
    which each map to properties on the underlying TimeSeries objects. It also contains a
    time vector corresponding to the time values that appear or will appear in the spreadsheet.

    Note that the units are stored within the TimeSeries objects, which means that they can
    are able to differ across rows.

    :param name: The name/title for this table
    :param tvec: Specify the time values for this table. All TimeSeries in the ts dict should have corresponding time values
    :param ts: Optionally specify an odict() of TimeSeries objects populating the rows. Could be populated after
    :param allowed_units: Optionally specify a list of allowed units that will appear as a dropdown
    :param comment: Optionally specify descriptive text that will be added as a comment to the name cell

    """

    def __init__(self, name, tvec: np.array = None, ts=None, allowed_units: list = None, comment: str = None, pop_type: str = None, default_all: bool = False):

        if ts is None:
            ts = sc.odict()

        self.name = name  #: Name for th quantity printed in Excel
        self.comment = comment  #: A comment that will be added in Excel
        self.tvec = [] if tvec is None else tvec  #: time axis (e.g. np.arange(2000,2019)) - all TimeSeries time values must exactly match one of the values here
        self.ts = ts  # : dict of :class:`TimeSeries` objects
        self.allowed_units = [x.title() if x in FS.STANDARD_UNITS else x for x in allowed_units] if allowed_units is not None else None  # Otherwise, can be an odict with keys corresponding to ts - leave as None for no restriction
        self.pop_type = pop_type

        self.ts_attributes = {}  #: Dictionary containing extra attributes to write along with each TimeSeries object.
        self.ts_attributes["Provenance"] = {}  # Include provenance attribute by default

        #  Keys are attribute name, values can be either a scalar or a dict keyed by the same keys as self.ts. Compared to units, uncertainty etc.
        #  attributes are store in the TDVE rather than in the TimeSeries
        self.assumption_heading = "Constant"  #: Heading to use for assumption column

        self.write_units = None  #: Write a column for units (if None, units will be written if any of the TimeSeries have units)
        self.write_uncertainty = None  #: Write a column for uncertainty (if None, uncertainty will be written if any of the TimeSeries have uncertainty)
        self.write_assumption = None  #: Write a column for assumption/constant (if None, assumption will be written if any of the TimeSeries have an assumption)

        self.default_all = default_all  #: Record whether the framework specifies that this TDVE should default to having an 'All' row instead of population-specific rows (the user can manually modify further)

    def __repr__(self):
        output = sc.prepr(self)
        return output

    @property
    def has_data(self) -> bool:
        """
        Check whether all time series have data entered

        :return: True if all of the TimeSeries objects stored in the TDVE have data
        """

        return all([x.has_data for x in self.ts.values()])

    @classmethod
    def from_rows(cls, rows: list):
        """
        Create new instance from Excel rows

        Given a set of openpyxl rows, instantiate a :class:`TimeDependentValuesEntry` object
        That is, the parent object e.g. :class:`ProjectData` is responsible for finding where the TDVE table is,
        and reading all of the rows associated with it (skipping ``#ignored`` rows) and then passing those rows,
        unparsed, to this function

        Headings for 'units', 'uncertainty', and 'assumption'/'constant' are optional and will be read in
        if they are present in the spreadsheet.

        :param rows: A list of rows
        :return: A new :class:`TimeDependentValuesEntry` instance

        """

        from .utils import TimeSeries  # Import here to avoid circular reference

        # Retrieve the name
        name = rows[0][0].value
        if name is None:
            raise Exception('The name of the table is missing. This can also happen if extra rows have been added without a "#ignore" entry in the first column')
        elif not sc.isstring(name):
            raise Exception("In cell %s of the spreadsheet, the name of the quantity assigned to this table needs to be a string" % rows[0][0].coordinate)
        name = name.strip()  # The name needs to be written back in a case sensitive form

        tdve = cls(name)
        known_headings = {"units", "uncertainty", "constant", "assumption"}
        headings, times, tdve.tvec = _parse_ts_header(rows[0], known_headings, skip_first=True)

        # Validate and process headings
        if not times and "constant" not in headings:
            raise Exception("Could not find an assumption or time-specific value - all tables must contain at least one of these values")
        tdve.write_units = True if "units" in headings else None
        tdve.write_uncertainty = True if "uncertainty" in headings else None
        tdve.write_assumption = True if "constant" in headings else None
        if "assumption" in headings:
            tdve.write_assumption = True
            tdve.assumption_heading = "Assumption"
        for heading in headings:
            if heading not in known_headings:
                # If it's not a known heading and it's a string, then it must be an attribute
                # Note that the way `headings` is populated by skipping i=0 ensures that the table name
                # is not interpreted as a heading
                tdve.ts_attributes[heading] = {}
        ts_entries = sc.odict()

        for row in rows[1:]:
            if row[0].value is None:
                continue

            if not row[0].data_type in {"s", "str"}:
                raise Exception("In cell %s of the spreadsheet, the name of the entry was expected to be a string, but it was not. The left-most column is expected to be a name. If you are certain the value is correct, add an single quote character at the start of the cell to ensure it remains as text" % row[0].coordinate)
            series_name = row[0].value.strip()

            if "units" in headings:
                units = cell_get_string(row[headings["units"]], allow_empty=True)
                if units is not None and units.lower().strip() in FS.STANDARD_UNITS:
                    units = units.lower().strip()  # Only lower and strip units if they are standard units
            else:
                units = None
            ts = TimeSeries(units=units)

            if "uncertainty" in headings:
                ts.sigma = cell_get_number(row[headings["uncertainty"]])
            else:
                ts.sigma = None

            if "constant" in headings:
                ts.assumption = cell_get_number(row[headings["constant"]])
            elif "assumption" in headings:
                ts.assumption = cell_get_number(row[headings["assumption"]])
            else:
                ts.assumption = None

            for attribute in tdve.ts_attributes:
                if attribute in headings:
                    # If it's a default attribute e.g. provenance, and it is missing from the databook, then don't populate it
                    tdve.ts_attributes[attribute][series_name] = row[headings[attribute]].value

            for t, idx in times.items():
                ts.insert(t, cell_get_number(row[idx]))  # If cell_get_number returns None, this gets handled accordingly by ts.insert()
            ts_entries[series_name] = ts

        tdve.ts = ts_entries
        return tdve

    def write(self, worksheet, start_row, formats, references: dict = None, widths: dict = None) -> int:
        """
        Write to cells in a worksheet

        Note that the year columns are drawn from the ``tvec`` attribute. To suppress the year columns (e.g. for the user to enter only an assumption)
        then set ``tvec`` to an empty array/list.

        :param worksheet: An xlsxwriter worksheet instance
        :param start_row: The first row in which to write values
        :param formats: Format dict for the opened workbook - typically the return value of :func:`standard_formats` when the workbook was opened
        :param references: References dict containing cell references for strings in the current workbook
        :param widths: ``dict`` storing column widths
        :return: The row index for the next available row for writing in the spreadsheet

        """

        assert self.assumption_heading in {"Constant", "Assumption"}, "Unsupported assumption heading"

        write_units = self.write_units if self.write_units is not None else any((ts.units is not None for ts in self.ts.values()))
        write_uncertainty = self.write_uncertainty if self.write_uncertainty is not None else any((ts.sigma is not None for ts in self.ts.values()))
        write_assumption = self.write_assumption if self.write_assumption is not None else any((ts.assumption is not None for ts in self.ts.values()))

        if not references:
            references = dict()

        current_row = start_row

        # First, assemble and write the headings
        headings = []
        headings.append(self.name)
        offset = 1  # This is the column where the time values start (after the 'or')

        # Next allocate attributes
        attribute_index = {}
        for attribute in self.ts_attributes:
            attribute_index[attribute] = offset
            headings.append(attribute)
            offset += 1

        if write_units:
            headings.append("Units")
            units_index = offset  # Column to write the units in
            offset += 1

        if write_uncertainty:
            headings.append("Uncertainty")
            uncertainty_index = offset  # Column to write the units in
            offset += 1

        if write_assumption:
            headings.append(self.assumption_heading)
            headings.append("")
            constant_index = offset
            offset += 2

        headings += [float(x) for x in self.tvec]
        for i, entry in enumerate(headings):
            worksheet.write(current_row, i, entry, formats["center_bold"])
            update_widths(widths, i, entry)

        if not pd.isna(self.comment):
            worksheet.write_comment(xlrc(current_row, 0), self.comment)

        # Now, write the TimeSeries objects - self.ts is an odict and whatever pops are present will be written in whatever order they are in
        for row_name, row_ts in self.ts.items():
            current_row += 1

            # Write the name
            if row_name in references:
                worksheet.write_formula(current_row, 0, references[row_name], formats["center_bold"], value=row_name)
                update_widths(widths, 0, row_name)
            else:
                worksheet.write_string(current_row, 0, row_name, formats["center_bold"])
                update_widths(widths, 0, row_name)

            # Write the attributes
            for attribute in self.ts_attributes:
                if isinstance(self.ts_attributes[attribute], dict):
                    if row_name in self.ts_attributes[attribute]:
                        val = self.ts_attributes[attribute][row_name]
                    else:
                        val = None
                else:
                    val = self.ts_attributes[attribute]

                if val is not None:
                    worksheet.write(current_row, attribute_index[attribute], val)
                    update_widths(widths, attribute_index[attribute], val)

            # Write the units
            if write_units:
                if row_ts.units:
                    if row_ts.units.lower().strip() in FS.STANDARD_UNITS:  # Preserve case if nonstandard unit
                        unit = row_ts.units.title().strip()
                    else:
                        unit = row_ts.units.strip()
                    worksheet.write(current_row, units_index, unit)
                    update_widths(widths, units_index, unit)
                else:
                    worksheet.write(current_row, units_index, FS.DEFAULT_SYMBOL_INAPPLICABLE)

                if self.allowed_units and isinstance(self.allowed_units, dict) and row_name in self.allowed_units:  # Add dropdown selection if there is more than one valid choice for the units
                    allowed = self.allowed_units[row_name]
                elif self.allowed_units and not isinstance(self.allowed_units, dict):
                    allowed = self.allowed_units
                else:
                    allowed = None

                if allowed:
                    worksheet.data_validation(xlrc(current_row, units_index), {"validate": "list", "source": allowed})

            if write_uncertainty:
                if row_ts.sigma is None:
                    worksheet.write(current_row, uncertainty_index, row_ts.sigma, formats["not_required"])  # NB. For now, uncertainty is always optional
                else:
                    worksheet.write(current_row, uncertainty_index, row_ts.sigma, formats["not_required"])

            if row_ts.has_data:
                format = formats["not_required"]
            else:
                format = formats["unlocked"]

            if write_assumption:
                worksheet.write(current_row, constant_index, row_ts.assumption, format)
                if len(self.tvec):
                    worksheet.write(current_row, constant_index + 1, "OR", formats["center"])
                    update_widths(widths, constant_index + 1, "OR")

            # Write the time values if they are present
            if len(self.tvec):
                content = [None] * len(self.tvec)  # Initialize an empty entry for every time in the TDVE's tvec

                for t, v in zip(row_ts.t, row_ts.vals):
                    # If the TimeSeries contains data for that time point, then insert it now
                    idx = np.where(self.tvec == t)[0]
                    if len(idx):
                        content[idx[0]] = v

                for idx, v in enumerate(content):
                    if v is None:
                        worksheet.write_blank(current_row, offset + idx, v, format)
                    else:
                        worksheet.write(current_row, offset + idx, v, format)
                    widths[offset + idx] = max(widths[offset + idx], 7) if offset + idx in widths else 7

                if write_assumption:
                    # Conditional formatting for the assumption
                    # Do this here, because after the loop above, we have easy and clear access to the range of cells to include in the formula
                    fcn_empty_times = 'COUNTIF(%s:%s,"<>" & "")>0' % (xlrc(current_row, offset), xlrc(current_row, offset + len(content) - 1))
                    # Hatched out if the cell will be ignored
                    worksheet.conditional_format(xlrc(current_row, constant_index), {"type": "formula", "criteria": "=" + fcn_empty_times, "format": formats["ignored"]})
                    worksheet.conditional_format(xlrc(current_row, constant_index), {"type": "formula", "criteria": "=AND(%s,NOT(ISBLANK(%s)))" % (fcn_empty_times, xlrc(current_row, constant_index)), "format": formats["ignored_warning"]})

        return current_row + 2  # Add two so there is a blank line after this table


def _parse_ts_header(row: list, known_headings: list, skip_first=False) -> Tuple[dict, dict, np.array]:
    """
    Internal function to read TDVE/TDC timeseries tables

    :param row: A list of openpyxl Cell instances corresponding to the heading row (the one containing the
                 headers for 'constant', 'uncertainty', 'provenance' etc., and the optional year values
    :param known_headings: A list of known headings that should be lowered for further processing. These
                           would correspond to the standard/default headings
    :param skip_first: If True, the first heading cell will be skipped. This is required for TDVE tables
                       where the top left cell is the name of the variable and not a column heading
    :return: A tuple containing
                - A dictionary of heading values (from string cell types) with {heading: column_index}.
                  Any known headings will have had `str.lower()` called on them.
                - A dictionary of time values and the columns they appear in. Datetimes will be converted
                  into floats at this step (so for example a spreadsheet containing 1/1/2022 will have
                  that converted to 2022.0 at this step)
                - A numpy array with sorted time values, that can be used as a time vector

    """

    headings = {}
    times = {}
    for i, cell in enumerate(row):
        v = cell.value
        if v is None or (i == 0 and skip_first):
            continue
        elif cell.data_type in {"s", "str"}:
            v = v.strip()
            if v.startswith("#ignore"):
                break
            elif v.lower() in known_headings:
                if v.lower() in headings:
                    raise Exception(f"Duplicate heading in cell {cell.coordinate} - another cell in the same row already has the same header '{v}' (possibly in cell {get_column_letter(1+headings[v.lower()])}{cell.row})")
                headings[v.lower()] = i
            else:
                if v in headings:
                    raise Exception(f"Duplicate heading in cell {cell.coordinate} - another cell in the same row already has the same header '{v}' (possibly in cell {get_column_letter(1+headings[v.lower()])}{cell.row})")
                headings[v] = i
        elif cell.is_date:
            year = datetime_to_year(v)
            if year in times:
                raise Exception(f"Duplicate year in cell {cell.coordinate} - another cell in the same row already contains the same year ({year}) (possibly in cell {get_column_letter(1+times[year])}{cell.row})")
            times[datetime_to_year(v)] = i
        elif cell.data_type == "n":
            if v in times:
                raise Exception(f"Duplicate year in cell {cell.coordinate} - another cell in the same row already contains the same year ({v}) (possibly in cell {get_column_letter(1+times[v])}{cell.row})")
            times[v] = i
        else:
            raise Exception("Unknown data type in cell %s of the spreadsheet - quantity must be a string or a number" % cell.coordinate)
    return headings, times, np.array(sorted(times), dtype=float)


def cell_get_string(cell, allow_empty=False) -> str:
    """
    Return string value from cell

    This function checks if a cell contains a string. If it does, the stripped value
    will be returned. Otherwise, an informative error will be raised

    Note that the string type is determined from the cell's value rather than
    the openpyxl cell data type.

    :param cell: An openpyxl cell
    :return: A string with the contents of the cell
    """
    if cell.value is None and allow_empty:
        return None
    elif not sc.isstring(cell.value):
        raise Exception("Cell %s needs to contain a string (i.e. not a number, date, or other cell type)" % cell.coordinate)
    else:
        return cell.value.strip()


def cell_get_number(cell, dtype=float):
    """
    Return numeric value from cell

    This function is to guard against accidentally having the Excel cell contain a string
    instead of a number. If a string has been entered, an error will be raised. The added value
    from this function is that if the Excel cell type is empty but the value is
    empty or ``N.A.`` then the value will be treated as though the cell was correctly set to a
    numeric type but had been left empty.

    The output is cast to ``dtype`` which means that code that requires numeric input from Excel
    can use this input to guarantee that the resulting number is of the correct type, or ``None``.

    :param cell: An openpyxl cell
    :param dtype: If the cell is numeric, cast to this type (default is `float` but could be `int` for example)
    :return: A scalar instance of ``dtype`` (e.g. ``float``) or ``None`` if cell is empty or being treated as empty
    :raises: :class:`Exception` if the cell contains a string

    """

    if cell.value is None:
        return None
    elif cell.data_type == "n":  # Numeric type
        return dtype(cell.value)
    elif cell.data_type == "s":  # Only do relatively expensive string processing if it's actually a string type
        s = cell.value.lower().strip()
        if s == FS.DEFAULT_SYMBOL_INAPPLICABLE:
            return None
        elif not s.replace("-", ""):
            return None

    raise Exception("Cell %s needs to contain a number" % cell.coordinate)


def validate_category(workbook, expected_category) -> None:
    """
    Check Atomica workbook type

    This function makes sure that a workbook has a particular category property
    stored within it, and displays an appropriate error message if not. If the
    category isn't present or doesn't start with 'atomica', just ignore it for
    robustness (instead, a parsing error will likely be raised)

    :param workbook: An openpyxl workbook
    :param category: The expected string category
    :raises: :class:`Exception` if the workbook category is not valid

    """

    category = workbook.properties.category
    if category and sc.isstring(category) and category.startswith("atomica:"):
        if category.strip() != expected_category.strip():
            expected_type = expected_category.split(":")[1].title()
            actual_type = category.split(":")[1].title()
            message = "Error loading %s - the provided file was a %s file" % (expected_type, actual_type)
            raise Exception(message)



"""
Note, need to run testworkflow.py first to generate the required files
"""

# Imports
import atomica as at
import os

# Setup
plot_initial = True

test = "sir"  # This test workflow only works with the SIR model

testdir = at.parent_dir()
tmpdir = testdir / "temp"

F = at.ProjectFramework(at.LIBRARY_PATH / test + "_framework.xlsx")
P = at.Project(name=test.upper() + " project", framework=F, do_run=False)
P.load_databook(databook_path=at.LIBRARY_PATH / test + "_databook.xlsx", make_default_parset=True, do_run=True)

at.export_results(P.results[0], tmpdir / test.upper() + " results")

P.save(tmpdir / test + ".prj")
P = at.Project.load(tmpdir / test + ".prj")

if plot_initial:

    for var in ["sus", "inf", "rec", "dead", "ch_all", "foi"]:
        P.results[0].get_variable(var, "adults")[0].plot()

    # Plot decomposition of population
    d = at.PlotData(P.results[0], outputs=["sus", "inf", "rec", "dead"], project=P)
    at.plot_series(d, plot_type="stacked")

    # Bar plot showing deaths, disaggregated by source compartment
    d = at.PlotData(P.results[0], outputs=["sus:dead", "inf:dead", "rec:dead"], t_bins=10, project=P)
    at.plot_bars(d, outer="results", stack_outputs="all")

    # Aggregate flows
    d = at.PlotData(P.results[0], outputs=[{"Death rate": ["infdeath:flow", "susdeath:flow"]}], project=P)
    at.plot_series(d)

    # Demonstrate how susdeath:flow sums over all of the tags sharing that label
    d = at.PlotData(P.results[0], outputs=["susdeath:flow"], project=P)
    at.plot_series(d)

    # Summing over the transitions between compartments in susdeath gives the same result
    d = at.PlotData(P.results[0], outputs=["sus:dead", "rec:dead"], project=P)
    at.plot_series(d, plot_type="stacked")

    # Summing via a function aggregation does the same thing
    d = at.PlotData([P.results[0]], outputs={"Susdeaths": "sus:dead+rec:dead"})
    at.plot_series(d)


# Make a scenario e.g. decreased infection rate
scvalues = dict()
scvalues["infdeath"] = dict()
scvalues["infdeath"]["adults"] = dict()
scvalues["infdeath"]["adults"]["y"] = [0.125]
scvalues["infdeath"]["adults"]["t"] = [2015.0]

scvalues["infdeath"]["adults"]["y"] = [0.125, 0.5]
scvalues["infdeath"]["adults"]["t"] = [2015.0, 2020.0]
scvalues["infdeath"]["adults"]["smooth_onset"] = [2, 3]

scvalues["infdeath"]["adults"]["y"] = [0.125, 0.25, 0.50, 0.50]
scvalues["infdeath"]["adults"]["t"] = [2015.0, 2020.0, 2025.0, 2030.0]
scvalues["infdeath"]["adults"]["smooth_onset"] = [4.0, 3.0, 2.0, 1.0]

s = at.ParameterScenario("increased_infections", scvalues)
s.run(project=P, parset=P.parsets["default"])

d = at.PlotData(P.results, outputs=["infdeath"])
at.plot_series(d, axis="results")

d = at.PlotData(P.results, outputs=["inf"])
at.plot_series(d, axis="results")

d = at.PlotData(P.results, outputs=["dead"])
at.plot_series(d, axis="results")

# plt.show()

# Synthesize the calibration target
# P.parsets['calibration_target'] = dcp(P.parsets[0])
# P.parsets['calibration_target'].name = 'calibration_target'
# par = P.parsets['calibration_target'].get_par('transpercontact')
# par.y_factor['adults']=0.2
# r2 = P.run_sim(parset='calibration_target')
# d = PlotData([P.results[0],r2], outputs=['ch_prev'])
# plot_series(d, axis='results',data=P.data)


# Perform calibration to get a calibrated parset
pars_to_adjust = [("transpercontact", "adults", 0.1, 1.9)]
output_quantities = []
for pop in P.parsets[0].pop_names:
    output_quantities.append(("ch_prev", pop, 1.0, "fractional"))
calibrated_parset = at.calibrate(P, P.parsets[0], pars_to_adjust, output_quantities, max_time=30)

# Plot the results before and after calibration
calibrated_results = P.run_sim(calibrated_parset)
d = at.PlotData([P.results[0], calibrated_results], outputs=["ch_prev"])
at.plot_series(d, axis="results", data=P.data)



import numpy as np
import atomica as at

testdir = at.parent_dir()
tmpdir = testdir / "temp"


def test_save():

    P = at.demo("sir")
    P.save(tmpdir / "test_project_save.prj")
    P2 = at.Project.load(tmpdir / "test_project_save.prj")
    P2.run_sim()


if __name__ == "__main__":
    test_save()



"""
Define classes for handling scenarios


This module implements the classes used to represent the four
main types of scenarios in Atomica:

- Parameter scenarios
- Budget scenarios
- Capacity scenarios
- Coverage scenarios

There are broadly two kinds of scenario

- Parameter scenarios which involve modifying a :class:`ParameterSet` instance
- Program scenarios (budget, capacity, coverage) which involve modifying a :class:`ProgramInstructions` instance

"""

import numpy as np
import sciris as sc
from .system import logger
from .utils import NamedItem, TimeSeries
from .programs import ProgramInstructions, ProgramSet
from .parameters import ParameterSet
from .results import Result

__all__ = ["Scenario", "CombinedScenario", "BudgetScenario", "CoverageScenario", "ParameterScenario"]


class Scenario(NamedItem):
    """
    Base scenario class

    All Scenario objects share this type. It is a NamedItem that also has an
    ``active`` property which represents whether the scenario should be run
    as part of :meth:`project.Project.run_scenarios`

    The optional variables ``parsetname`` and ``progsetname`` reference the name of parsets
    and progsets to use via :meth:`Scenario.run`. These should match the names of objects in the
    project passed into :meth:`Scenario.run`. However, :meth:`Scenario.run` can also take in
    a parset and/or progset directly, allowing the scenario to be used with other parsets and
    progsets on the fly. If  ``parsetname`` and ``progsetname`` are not set in the :class:`Scenario`
    then they must be passed into :meth:`Scenario.run`.

    :param name: The name of the scenario - also sets the result name via :meth:`Scenario.run`
    :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project
    :param progsetname: If running via ``Project.run_scenarios`` this identifies which progset to use. If set to ``None`` then programs will not be used
    :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario

    """

    def __init__(self, name: str, active: bool = True, parsetname: str = None, progsetname: str = None):
        NamedItem.__init__(self, name)
        self.parsetname = parsetname  #: Specify parset name when run via ``Project.run_scenarios``
        self.progsetname = progsetname  #: Specify progset name when run via ``Project.run_scenarios``
        self.active = active  #: Flag whether the scenario should be run via ``Project.run_scenarios``

    def get_parset(self, parset, project) -> ParameterSet:
        """
        Get scenario parset

        If the derived scenario class modifies the parset, return the modified version

        :param parset: Input :class:`ParameterSet`
        :return: Modified parset for use in the simulation

        """

        return parset

    def get_progset(self, progset: ProgramSet, project) -> ProgramSet:
        """
        Get scenario progset

        If the derived scenario class modifies the progset, return the modified version

        :param progset: Input :class:`ProgramSet`
        :return: Modified progset for use in the simulation

        """

        return progset

    def get_instructions(self, progset: ProgramSet, project) -> ProgramInstructions:
        """
        Get scenario instructions

        If the derived scenario class produces  program instructions, return
        them here.

        :param progset: Input :class:`ProgramSet`
        :return: :class:`ProgramInstructions` instance, or None if no instructions (in which case, programs will not be used)

        """

        return None

    def run(self, project, parset: ParameterSet = None, progset: ProgramSet = None, store_results: bool = True) -> Result:
        """
        Run scenario

        :param project: A :class:`Project` instance
        :param parset: Optionally a :class:`ParameterSet` instance, otherwise will use ``self.parsetname``
        :param progset: Optionally a :class:`ProgramSet` instance, otherwise will use ``self.progsetname``
        :param store_results: If True, the results will be copied into the project
        :return: A :class:`Result` object

        """

        if parset is None:
            parset = project.parsets[self.parsetname]
        else:
            parset = project.parset(parset)

        if progset is None and self.progsetname is not None:
            progset = project.progsets[self.progsetname]
        elif progset is not None:
            progset = project.progset(progset)

        parset = self.get_parset(parset, project)
        progset = self.get_progset(progset, project)
        instructions = self.get_instructions(progset, project)

        if progset is not None:
            if instructions is None:
                raise Exception("If using programs, the scenario must contain instructions specifying at minimum the program start year")
            result = project.run_sim(parset=parset, progset=progset, progset_instructions=instructions, result_name=self.name, store_results=store_results)
        else:
            result = project.run_sim(parset=parset, result_name=self.name, store_results=store_results)

        return result


class CombinedScenario(Scenario):
    """
    Define combined (budget+program) scenario

    This object stores both a set of scenario values and a set of program instructions.
    This allows it to simultaneously apply parameter, budget, and coverage overwrites.

    As usual, parameter values from programs take precedence over parameter values from parsets, and
    within programs, coverage takes precedence ovetoxr budgets.

    :param name: The name of the scenario. This will also be used to name the result
    :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario
    :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project
    :param progsetname: If running via ``Project.run_scenarios`` this identifies which progset to use. If set to ``None`` then programs will not be used
    :param scenario_values: Parameter value overwrites, used as input to :class:`ParameterScenario`
    :param instructions: A :class`ProgramInstructions` instance containing required program overwrites (budget, capacity, coverage)

    """

    def __init__(self, name: str = None, active: bool = True, parsetname: str = None, progsetname: str = None, scenario_values: dict = None, instructions: ProgramInstructions = None, interpolation: str = "linear"):
        super().__init__(name, active, parsetname, progsetname)
        self.scenario_values = scenario_values  #: Parameter scenario values (see :class:`ParameterScenario`)
        self.interpolation = interpolation  #: Interpolation method to use for parameter overwrite
        self.instructions = instructions  # : Program instructions for budget scenario (should already contain required overwrites)

    def get_parset(self, parset, project) -> ParameterSet:
        if self.scenario_values is not None:
            scenario_parset = ParameterScenario(scenario_values=self.scenario_values, interpolation=self.interpolation).get_parset(parset, project)
        else:
            scenario_parset = parset
        return scenario_parset

    def get_instructions(self, progset: ProgramSet, project) -> ProgramInstructions:
        return self.instructions


class BudgetScenario(Scenario):
    def __init__(self, name=None, active: bool = True, parsetname: str = None, progsetname: str = None, alloc: dict = None, start_year=2019):
        super().__init__(name, active, parsetname, progsetname)
        self.start_year = start_year  # Program start year
        self.alloc = sc.dcp(alloc) if alloc is not None else sc.odict()

    def get_instructions(self, progset: ProgramSet, project) -> ProgramInstructions:
        return ProgramInstructions(start_year=self.start_year, alloc=self.alloc)


class CoverageScenario(Scenario):
    def __init__(self, name=None, active: bool = True, parsetname: str = None, progsetname: str = None, coverage: dict = None, start_year=2019):
        super().__init__(name, active, parsetname, progsetname)
        self.start_year = start_year  # Program start year
        self.coverage = sc.dcp(coverage) if coverage is not None else sc.odict()

    def get_instructions(self, progset: ProgramSet, project) -> ProgramInstructions:
        return ProgramInstructions(start_year=self.start_year, coverage=self.coverage)


class ParameterScenario(Scenario):
    """
    Define and run parameter scenarios

    This object stores overwrites to parameter values that are used to modify
    a :class:`ParameterSet` instance before running a simulation.

    Example usage:

    >>> scvalues = dict()
    >>> param = 'birth_transit'
    >>> scvalues[param] = dict()
    >>> scvalues[param]['Pop1'] = dict()
    >>> scvalues[param]['Pop1']['y'] = [3e6, 1e4, 1e4, 2e6]
    >>> scvalues[param]['Pop1']['t'] = [2003.,2004.,2014.,2015.]
    >>> pscenario = ParameterScenario(name="examplePS",scenario_values=scvalues)

    :param name: The name of the scenario. This will also be used to name the result
    :param scenario_values: A dict of overwrites to parameter values. The structure is
        ``{parameter_label: {pop_identifier: dict o}`` where the overwrite ``o`` contains keys
         - ``t`` : np.array or list with year values
         - ``y`` : np.array or list with corresponding parameter values
         The ``pop_identifier`` is a single population name if the ``parameter_label`` corresponds to a Framework parameter
         or it should be a tuple/list ``(from_pop,to_pop)`` if the ``parameter_label`` corresponds to an interaction or transfer.
    :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario
    :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project
    :param interpolation: The specified interpolation method will be used to interpolate scenario values onto simulation times. Common options are 'linear' (smoothly change) and 'previous' (stepped)

    """

    def __init__(self, name: str = None, scenario_values: dict = None, active: bool = True, parsetname: str = None, interpolation: str = "linear"):

        super().__init__(name, active, parsetname)
        self.scenario_values = sc.dcp(scenario_values) if scenario_values is not None else dict()  #: Store dictionary containing the overwrite values
        self.interpolation = interpolation  #: Stores the name of a supported interpolation method

    def add(self, par_name: str, pop_name: str, t, y) -> None:
        """
        Add overwrite to scenario

        This method adds a TimeSeries with parameter overwrite values to the
        ParameterSet

        Example usage:

        >>> pscenario = ParameterScenario(name="examplePS")
        >>> pscenario.add('rec_rate','Pop1',[2004.,2014],[3e6, 1e4])
        >>> pscenario.add('aging',('Pop1','Pop2'),[2004.,2014],[0.05,0.06])

        This can provide a more readable way to define a parameter scenario, without having to
        assemble a dict of the overwrites in advance.

        :param par_name: Name of the parameter to overwrite. This can also be the code name of a transfer or interaction
        :param pop_name: Population to overwrite values for. If the ``par_name`` is a transfer or interaction, this
                         should be a tuple containing ``(from_pop,to_pop)``
        :param t: scalar, list, or array of times
        :param y: scalar, list, or array of overwrite values
        :param end_overwrite: If True, after the final overwrite, the parameter will revert to its baseline value

        """

        t = sc.promotetoarray(t).copy()
        y = sc.promotetoarray(y).copy()

        assert len(t) == len(y), "To add an overwrite, the same number of time points and values must be provided"

        if par_name not in self.scenario_values:
            self.scenario_values[par_name] = dict()
        if pop_name not in self.scenario_values[par_name]:
            self.scenario_values[par_name][pop_name] = dict()
        self.scenario_values[par_name][pop_name] = {"t": t, "y": y}

    def get_parset(self, parset: ParameterSet, project) -> ParameterSet:
        """
        Return modified parset

        This method takes in a :class:`ParameterSet` and modifies it by applying the overwrites
        present in the scenario. This can thus be used to return the :class:`ParameterSet` for use in
        other simulations that are manually run, or to do things like perform a budget scenario simulation
        in conjunction with a parameter scenario.

        The returned :class:`ParameterSet` will have been pre-interpolated onto the simulation times.

        :param parset: A :class:`ParameterSet` instance
        :param project: A :class:`Project` instance (required for simulation times and to identify function parameters)
        :return: A new :class:`ParameterSet` object

        """

        new_parset = sc.dcp(parset)
        new_parset.name = self.name + "_" + parset.name
        tvec = project.settings.tvec  # Simulation times

        for par_label in self.scenario_values.keys():

            if par_label in new_parset.pars:
                has_function = project.framework.pars.at[par_label, "function"]  # Flag whether this is a function parameter in the framework
            else:
                has_function = False  # Interactions and transfers do not have functions

            for pop_specifier, overwrite in self.scenario_values[par_label].items():

                if sc.isstring(pop_specifier):
                    par = new_parset.get_par(par_label)
                    pop_label = pop_specifier
                else:
                    par = new_parset.get_par(par_label, pop_specifier[0])
                    pop_label = pop_specifier[1]

                # Sanitize the overwrite values
                overwrite = sc.dcp(overwrite)
                overwrite["t"] = sc.promotetoarray(overwrite["t"]).astype("float")  # astype('float') converts None to np.nan
                overwrite["y"] = sc.promotetoarray(overwrite["y"]).astype("float")
                idx = ~np.isnan(overwrite["t"]) & ~np.isnan(overwrite["y"])
                if not np.any(idx):
                    continue

                # Expand out the baseline values and remove any other values
                scen_start = min(overwrite["t"])
                vals = par.interpolate(tvec[tvec < scen_start], pop_label)  # Interpolate parameter onto valid sim times, using default interpolation. To override this, interpolate the parameter before calling ``ParameterScenario.get_parset()``
                par.ts[pop_label].t = tvec[tvec < scen_start].tolist()
                par.ts[pop_label].vals = vals.tolist()

                # Insert the overwrites
                assert len(overwrite["t"]) == len(overwrite["y"]), "Number of time points in overwrite does not match number of values"
                for t, y in zip(overwrite["t"], overwrite["y"]):
                    par.ts[pop_label].insert(t, y)
                par.smooth(tvec[tvec >= scen_start], pop_names=pop_label, method=self.interpolation)

                # Disable parameter function during scenario
                if has_function:
                    par.skip_function[pop_label] = (scen_start, np.inf)

        return new_parset



# Test that parameter derivatives work
# Indirectly test that frameworks with no compartments work

import numpy as np
import atomica as at
import os


def test_derivative():

    testdir = at.parent_dir()  # Must be relative to current file to work with tox

    F = at.ProjectFramework(testdir / "framework_derivative_test.xlsx")
    D = at.ProjectData.new(F, np.arange(2000, 2010), pops={"mosquitos": "Mosquitos"}, transfers=0)

    P = at.Project(name="test", framework=F, do_run=False)
    P.load_databook(databook_path=D.to_spreadsheet(), make_default_parset=True, do_run=True)

    dm_prev = P.results[0].get_variable("dm_prev", "mosquitos")[0]
    m_prev = P.results[0].get_variable("m_prev", "mosquitos")[0]
    converge = P.results[0].get_variable("converge", "mosquitos")[0]

    assert np.allclose(dm_prev.vals[m_prev.t == 2000], 0.1, equal_nan=True)
    assert np.allclose(dm_prev.vals[m_prev.t == 2010], 0.1, equal_nan=True)

    assert np.allclose(m_prev.vals[m_prev.t == 2000], 0.5, equal_nan=True)
    assert np.allclose(m_prev.vals[m_prev.t == 2000.25], 0.5 + 0.1 * P.settings.sim_dt, equal_nan=True)
    assert np.allclose(m_prev.vals[m_prev.t == 2002.5], 0.75, equal_nan=True)
    assert np.allclose(m_prev.vals[m_prev.t == 2005], 1.0, equal_nan=True)
    assert np.allclose(m_prev.vals[m_prev.t == 2010], 1.0, equal_nan=True)

    print("Test successfully completed")

    assert np.allclose(converge.vals[-1], 1.0, equal_nan=True)  # Should converge to 1
    converge.plot()


if __name__ == "__main__":
    test_derivative()



"""
Version:
"""

import matplotlib.pyplot as plt
import numpy as np
import sciris as sc
import atomica as at
import logging

logger = logging.getLogger()

# Write the log to a file
# h = logging.FileHandler('testworkflow.log',mode='w')
# logger.addHandler(h)

# Setting DEBUG level before importing Atomica will display the structure warnings occurring during import
# logger.setLevel('DEBUG')


# Atomica has INFO level logging by default which is set when Atomica is imported, so need to change it after importing
# logger.setLevel('DEBUG')
test = "sir"  # Only really works for SIR


# np.seterr(all='raise')


def run_optimization(proj, optimization, instructions):
    unoptimized_result = proj.run_sim(parset=proj.parsets["default"], progset=proj.progsets["default"], progset_instructions=instructions, result_name="unoptimized")
    optimized_instructions = at.optimize(proj, optimization, parset=proj.parsets["default"], progset=proj.progsets["default"], instructions=instructions)
    optimized_result = proj.run_sim(parset=proj.parsets["default"], progset=proj.progsets["default"], progset_instructions=optimized_instructions, result_name="optimized")
    return unoptimized_result, optimized_result


# STANDARD OUTCOME OPTIMIZATION
# In this example, Treatment 2 is more effective than Treatment 1. The initial allocation has the budget
# mostly allocated to Treatment 1, and the result of optimization should be that the budget gets
# reallocated to Treatment 2


def test_standard():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = list()
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    d = at.PlotData([unoptimized_result, optimized_result], outputs=["ch_all"], project=P)
    at.plot_series(d, axis="results")


# UNRESOLVABLE CONSTRAINTS
# If the user specifies bounds on individual spending that are inconsistent with the
# total spending constraint, an informative error should be raised. This test verifies
# that this is detected correctly


def test_unresolvable():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 50.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending

    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 10.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 10.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint(t=2020, total_spend=201)  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    try:
        (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)
    except at.UnresolvableConstraint as e:
        print(e)
        print("Correctly raised UnresolvableConstraint error")

    constraints.total_spend = sc.promotetoarray(5)
    try:
        (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)
    except at.UnresolvableConstraint as e:
        print(e)
        print("Correctly raised UnresolvableConstraint error")


# STANDARD OUTCOME OPTIMIZATION, MINIMIZE DEATHS
# In this example, Treatment 2 is more effective than Treatment 1. The initial allocation has the budget
# mostly allocated to Treatment 1, and the result of optimization should be that the budget gets
# reallocated to Treatment 2


def test_standard_mindeaths():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))
    measurables = at.MinimizeMeasurable(":dead", 2030)
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    d = at.PlotData([unoptimized_result, optimized_result], outputs=[":dead"], project=P)
    at.plot_series(d, axis="results")


# DELAYED OUTCOME OPTIMIZATION
# In this example, Treatment 2 is more effective than Treatment 1. However, we are given the budget in
# 2020 and are only allowed to change it from 2025.


def test_delayed():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", at.TimeSeries([2020, 2024], [50, 50])), ("Treatment 2", at.TimeSeries([2020, 2024], [1, 1]))])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2025, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2025, "abs", 0.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)

    plt.figure()
    plt.plot(t, unoptimized_spending["Treatment 1"], label="Unoptimized Treatment 1")
    plt.plot(t, optimized_spending["Treatment 1"], label="Optimized Treatment 1")
    plt.plot(t, unoptimized_spending["Treatment 2"], label="Unoptimized Treatment 2")
    plt.plot(t, optimized_spending["Treatment 2"], label="Optimized Treatment 2")
    plt.legend()
    plt.title("Fixed spending in 2020, optimized from 2025 onwards")


# MULTI YEAR FIXED
# In this example (requested by applications) the total spend constraint
# is explicitly specified in several years, yielding a time-dependent scale-up
# Note that the total spend constraint necessarily only affects programs which are
# optimizable (i.e. if a program is excluded from the optimization, then it won't be
# touched by the total spending constraint). This also applies in the case where the
# different programs are optimizable in different years. The flip side of this is that
# typically the Optimization should therefore contain an adjustment in every year that
# the total spend is specified - which can be done simply by passing the array of times
# to the SpendingAdjustment constructor, as demonstrated here
#
# In this example, the optimal solution is to spend as much as possible on Program 2, subject
# to constraints. Thus, in 2020 the optimal budget is $5 on program 1 and $95 on program 2,
# and in 2040 the optimal budget is $25 on program 1 and $125 on program 2


def test_multiyear_fixed():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", at.TimeSeries([2020, 2024], [50, 50])), ("Treatment 2", at.TimeSeries([2020, 2024], [1, 1]))])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", [2020, 2024], "abs", 5.0, 100.0, 5.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", [2020, 2024], "abs", 5.0, 125.0, 5.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint(t=[2020, 2024], total_spend=[100, 150])  # Cap total spending in all years
    # Use PSO because this example seems a bit susceptible to local minima with ASD
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, method="pso")  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)
    print(unoptimized_spending)
    print(optimized_spending)

    d = at.PlotData.programs(optimized_result)
    at.plot_series(d, plot_type="stacked")
    plt.title("Scale up spending to 100 in 2020 and 150 in 2040")


# MULTI YEAR RELATIVE
# This is an interesting example. The total budget in 2020 is fixed at $50
# Then, we have adjustments in 2022 and 2024 with total spending constraints of
# $75 and $150 respectively. The spending is fixed until 2022, so treatment 1
# remains at $49 until then. In 2022, treatment 1 should be defunded to its minimum, so
# we have $5 on treatment 1, and $70 on treatment 2. Then, in 2024, we should max out
# treatment 2 at $100, and allocate $50 to treatment 1 again.
# Note how the spending is linearly ramped in between the times when spending is specified,
# whether explicitly in the allocation or through the spending adjustment


def test_multiyear_relative():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", at.TimeSeries([2020], [49])), ("Treatment 2", at.TimeSeries([2020], [1]))])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", [2022, 2024], "abs", 1.0, 100))
    adjustments.append(at.SpendingAdjustment("Treatment 2", [2022, 2024], "abs", 1.0, 100))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint(t=[2022, 2024], budget_factor=[1.5, 3.0])  # Cap total spending in all years
    # Use PSO because this example seems a bit susceptible to local minima with ASD
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, method="pso")  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)
    print(unoptimized_spending)
    print(optimized_spending)

    d = at.PlotData.programs(optimized_result)
    at.plot_series(d, plot_type="stacked")
    plt.title("Scale up spending to 1x in 2020 and 1.5x in 2040")


# GRADUAL OUTCOME OPTIMIZATION
# This is similar to ramped constraints, except the constraint is time-based rather than
# rate-of-change based. That is, rather than limiting the amount by which the spending can
# change, it is specified that the total change takes place over a certain number of years.
# In this case, we have specified spending in 2020 and want to meet spending targets in 2025
# with the caveat that the rollout of the change will take 3 years. Therefore, we fix the spending
# in 2022 and apply the adjustment in 2025, resulting in a smooth change in spending from 2022-2025


def test_gradual():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", at.TimeSeries([2020, 2022], [50, 50])), ("Treatment 2", at.TimeSeries([2020, 2022], [1, 1]))])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2025, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2025, "abs", 0.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)

    plt.figure()
    plt.plot(t, unoptimized_spending["Treatment 1"], label="Unoptimized Treatment 1")
    plt.plot(t, optimized_spending["Treatment 1"], label="Optimized Treatment 1")
    plt.plot(t, unoptimized_spending["Treatment 2"], label="Unoptimized Treatment 2")
    plt.plot(t, optimized_spending["Treatment 2"], label="Optimized Treatment 2")
    plt.legend()
    plt.title("Fixed spending in 2020, optimized in 2025 with 3y onset")


# MIXED TIMING OUTCOME OPTIMIZATION
# Treatment 2 can have its funding adjusted in both 2023 and 2027. Treatment 1 can have its funding
# only adjusted in 2023. The default allocation sees the funding smoothly change to equal funding on both
# programs in 2022. The total spend constraint is only applied in 2023, at which time both Treatment 1 and
# Treatment 2 are adjustable and spending is capped at 25+25=50 units. In 2027, Treatment 2 is the only
# adjustable program and there is no total spending constraint. Thus spending on this program should hit
# the upper bound. Thus the optimal spending pattern will be spending $50 on Treatment 2 in 2023, and $100
# on Treatment 2 in 2025, with $0 spend on Treatment 1 from 2023 onwards


def test_mixed_timing():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", at.TimeSeries([2020, 2022], [50, 25])), ("Treatment 2", at.TimeSeries([2020, 2022], [1, 25]))])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2023, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", [2023, 2027], "abs", 0.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint(t=2023)  # Cap total spending in 2023 only
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)

    plt.figure()
    plt.plot(t, unoptimized_spending["Treatment 1"], label="Unoptimized Treatment 1")
    plt.plot(t, optimized_spending["Treatment 1"], label="Optimized Treatment 1")
    plt.plot(t, unoptimized_spending["Treatment 2"], label="Unoptimized Treatment 2")
    plt.plot(t, optimized_spending["Treatment 2"], label="Optimized Treatment 2")
    plt.legend()
    plt.title("Multi-time optimization in 2023 and 2025 (constrained in 2023)")


def test_parametric_paired():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.PairedLinearSpendingAdjustment(["Treatment 1", "Treatment 2"], [2020, 2025]))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = None  # Total spending constraint is automatically satisfied by the paired parametric adjustment
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    t = optimized_result.model.t
    unoptimized_spending = unoptimized_result.model.progset.get_alloc(t, unoptimized_result.model.program_instructions)
    optimized_spending = optimized_result.model.progset.get_alloc(t, optimized_result.model.program_instructions)

    plt.figure()
    plt.plot(t, unoptimized_spending["Treatment 1"], label="Unoptimized Treatment 1")
    plt.plot(t, optimized_spending["Treatment 1"], label="Optimized Treatment 1")
    plt.plot(t, unoptimized_spending["Treatment 2"], label="Unoptimized Treatment 2")
    plt.plot(t, optimized_spending["Treatment 2"], label="Optimized Treatment 2")
    plt.legend()
    plt.title("Paired linear funding transfer from 2020-2025")


# MONEY MINIMIZATION
# In this example, Treatment 2 is more effective than Treatment 1. If we spend
# $40 on Treatment 2, then there are 729.90 people alive in 2030. Since Treatment 2 is more
# effective, the cheapest way to achieve at least 729.90 people alive in 2030 is to
# spend only ~$40 on Treatment 2. So to do this optimization, we start by spending $100 on both
# Treatment 1 and Treatment 2 and demonstrate that money optimization where we minimize total
# spend subject to the constraint of the total people alive being at least 729.90 in 2030


def test_minmoney():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 100.0), ("Treatment 2", 100.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending

    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))  # We can adjust Treatment 1
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))  # We can adjust Treatment 2

    measurables = []
    measurables.append(at.AtLeastMeasurable("ch_all", 2030, 729.90))  # Need at least 729.90 people in 2030
    measurables.append(at.MinimizeMeasurable("Treatment 1", 2020))  # Minimize 2020 spending on Treatment 1
    measurables.append(at.MinimizeMeasurable("Treatment 2", 2020))  # Minimize 2020 spending on Treatment 2

    constraints = None  # No extra constraints aside from individual bounds

    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, method="pso")  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    d = at.PlotData([unoptimized_result, optimized_result], outputs=["ch_all"], project=P)
    at.plot_series(d, axis="results")


# MONEY MINIMIZATION 2
# Essentially the same example as ``test_minmoney``. However, in this example, we want to both minimize spending
# and increase the number of people alive by some percentage. If we spend about $1000 on Treatment 2, then we can
# achieve 731.5 people alive in 2030, compared to 729.95 if we spend $50 on each program. Thus, we set our initial
# budget as $50 on each program, and explicitly set the initial spend to $2000 on each program, and request a
# 0.2% increase in ``ch_all``. The optimization should spend close to $0 on Treatment 1 and about $1000 on Treatment 2


def test_minmoney_relative():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 50.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending. Note that this does NOT have the scaled up initial spends for money minimization

    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 2000.0, initial=2000.0))  # We can adjust Treatment 1
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 2000.0, initial=2000.0))  # We can adjust Treatment 2

    measurables = []
    measurables.append(at.IncreaseByMeasurable("ch_all", 2030, 0.002))  # Get a 0.02% increase in ch_all in 2030
    measurables.append(at.MinimizeMeasurable("Treatment 1", 2020))  # Minimize 2020 spending on Treatment 1
    measurables.append(at.MinimizeMeasurable("Treatment 2", 2020))  # Minimize 2020 spending on Treatment 2

    constraints = None  # No extra constraints aside from individual bounds

    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, method="pso")  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    d = at.PlotData([unoptimized_result, optimized_result], outputs=["ch_all"], project=P)
    at.plot_series(d, axis="results")
    print("Change: %.4f" % (d.series[1].vals[-1] / d.series[0].vals[-1]))


# MONEY MINIMIZATION 3
# Essentially the same example as ``test_minmoney_relative`` but now we want an absolute increase
# of 1 person alive


def test_minmoney_absolute():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 50.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending. Note that this does NOT have the scaled up initial spends for money minimization

    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 2000.0, initial=2000.0))  # We can adjust Treatment 1
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 2000.0, initial=2000.0))  # We can adjust Treatment 2

    measurables = []
    measurables.append(at.IncreaseByMeasurable("ch_all", 2030, 1, target_type="abs"))  # Get a 0.02% increase in ch_all in 2030
    measurables.append(at.MinimizeMeasurable("Treatment 1", 2020))  # Minimize 2020 spending on Treatment 1
    measurables.append(at.MinimizeMeasurable("Treatment 2", 2020))  # Minimize 2020 spending on Treatment 2

    constraints = None  # No extra constraints aside from individual bounds

    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints, method="pso")  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    d = at.PlotData([unoptimized_result, optimized_result], outputs=["ch_all"], project=P)
    at.plot_series(d, axis="results")
    print("Before: %.2f, After: %.2f" % (d.series[0].vals[-1], d.series[1].vals[-1]))


def test_cascade_final_stage():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    # This is the same as the 'standard' example, just setting up the fact that we can adjust spending on Treatment 1 and Treatment 2
    # and want a total spending constraint
    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years

    # CASCADE MEASURABLE
    # This measurable will maximize the number of people in the final cascade stage, whatever it is
    measurables = at.MaximizeCascadeStage("main", [2030], pop_names="all")  # NB. make sure the objective year is later than the program start year, otherwise no time for any changes

    # This is the same as the 'standard' example, just running the optimization and comparing the results
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)
    unoptimized_result = P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=instructions, result_name="unoptimized")
    optimized_instructions = at.optimize(P, optimization, parset=P.parsets["default"], progset=P.progsets["default"], instructions=instructions)
    optimized_result = P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=optimized_instructions, result_name="optimized")

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    at.plot_multi_cascade([unoptimized_result, optimized_result], "main", pops="all", year=2020)


def test_cascade_multi_stage():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))

    # CASCADE MEASURABLE
    measurables = at.MaximizeCascadeStage(None, [2017, 2018], pop_names="all", cascade_stage=["Number ever infected", "Recovered"])  # NB. make sure the objective year is later than the program start year, otherwise no time for any changes
    # This is the same as the 'standard' example, just running the optimization and comparing the results
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables)
    P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=instructions, result_name="baseline")
    optimized_instructions = at.optimize(P, optimization, parset=P.parsets["default"], progset=P.progsets["default"], instructions=instructions)
    P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=optimized_instructions, result_name="optimized")


def test_cascade_conversions():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    # This is the same as the 'standard' example, just setting up the fact that we can adjust spending on Treatment 1 and Treatment 2
    # and want a total spending constraint

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = []
    adjustments.append(at.SpendingAdjustment("Treatment 1", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Treatment 2", 2020, "abs", 0.0, 100.0))
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years

    # CASCADE MEASURABLE
    # This measurable will be
    measurables = at.MaximizeCascadeConversionRate("main", [2030], pop_names="all")  # NB. make sure the objective year is later than the program start year, otherwise no time for any changes

    # This is the same as the 'standard' example, just running the optimization and comparing the results
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)
    unoptimized_result = P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=instructions, result_name="unoptimized")
    optimized_instructions = at.optimize(P, optimization, parset=P.parsets["default"], progset=P.progsets["default"], instructions=instructions)
    optimized_result = P.run_sim(parset=P.parsets["default"], progset=P.progsets["default"], progset_instructions=optimized_instructions, result_name="optimized")

    for adjustable in adjustments:
        print("%s - before=%.2f, after=%.2f" % (adjustable.name, unoptimized_result.model.program_instructions.alloc[adjustable.name].get(2020), optimized_result.model.program_instructions.alloc[adjustable.name].get(2020)))  # TODO - add time to alloc

    at.plot_cascade(unoptimized_result, "main", pops="all", year=2030)
    plt.title("Unoptimized")
    at.plot_cascade(optimized_result, "main", pops="all", year=2030)
    plt.title("Optimized")


def test_package_fixed():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 50.0), ("Harm reduction 1", 50.0), ("Harm reduction 2", 50.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = list()

    pkg = at.SpendingPackageAdjustment("TreatmentPackage", 2020, ["Treatment 1", "Treatment 2"], initial_spends=[50, 1], min_total_spend=51, max_total_spend=51)
    adjustments.append(pkg)
    adjustments.append(at.SpendingAdjustment("Risk avoidance", 2020, "abs", 0.0, 200.0))
    adjustments.append(at.SpendingAdjustment("Harm reduction 1", 2020, "abs", 0.0, 200.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for prog in alloc:
        print("%s - before=%.2f, after=%.2f" % (prog, unoptimized_result.model.program_instructions.alloc[prog].get(2020), optimized_result.model.program_instructions.alloc[prog].get(2020)))  # TODO - add time to alloc


def test_package_variable():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 0.0), ("Harm reduction 1", 0.0), ("Harm reduction 2", 0.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = list()

    pkg = at.SpendingPackageAdjustment("TreatmentPackage", 2020, ["Treatment 1", "Treatment 2"], initial_spends=[50, 1], min_total_spend=40, max_total_spend=60)
    adjustments.append(pkg)
    adjustments.append(at.SpendingAdjustment("Risk avoidance", 2020, "abs", 0.0, 100.0))
    adjustments.append(at.SpendingAdjustment("Harm reduction 1", 2020, "abs", 0.0, 100.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint(200, 2020)  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for prog in alloc:
        print("%s - before=%.2f, after=%.2f" % (prog, unoptimized_result.model.program_instructions.alloc[prog].get(2020), optimized_result.model.program_instructions.alloc[prog].get(2020)))  # TODO - add time to alloc

    # The check below should be satisfied, but since optimization is stochastic, it's not guaranteed to be met during automated testing
    # assert optimized_result.model.program_instructions.alloc["Treatment 2"].get(2020) > optimized_result.model.program_instructions.alloc["Treatment 1"].get(2020)


def test_package_fixed_prop():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 50.0), ("Harm reduction 1", 50.0), ("Harm reduction 2", 50.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = list()

    pkg = at.SpendingPackageAdjustment("TreatmentPackage", 2020, ["Treatment 1", "Treatment 2"], initial_spends=[10, 40], min_total_spend=40, max_total_spend=100, fix_props=True)
    assert len(pkg.adjustables) == 1
    adjustments.append(pkg)
    adjustments.append(at.SpendingAdjustment("Risk avoidance", 2020, "abs", 0.0, 200.0))
    adjustments.append(at.SpendingAdjustment("Harm reduction 1", 2020, "abs", 0.0, 200.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for prog in alloc:
        print("%s - before=%.2f, after=%.2f" % (prog, unoptimized_result.model.program_instructions.alloc[prog].get(2020), optimized_result.model.program_instructions.alloc[prog].get(2020)))  # TODO - add time to alloc

    # The package should have the full $100 allocated to it, but split in the same 10:40 ratio i.e. $80 on treatment 2 and $20 on treatment 1
    # assert np.isclose(optimized_result.model.program_instructions.alloc["Treatment 2"].get(2020), 80)
    # assert np.isclose(optimized_result.model.program_instructions.alloc["Treatment 1"].get(2020), 20)


def test_package_all_fixed():

    P = at.demo(which=test, do_run=False)
    P.update_settings(sim_end=2030.0)

    alloc = sc.odict([("Risk avoidance", 50.0), ("Harm reduction 1", 50.0), ("Harm reduction 2", 50.0), ("Treatment 1", 50.0), ("Treatment 2", 1.0)])

    instructions = at.ProgramInstructions(alloc=alloc, start_year=2020)  # Instructions for default spending
    adjustments = list()

    pkg = at.SpendingPackageAdjustment("TreatmentPackage", 2020, ["Treatment 1", "Treatment 2"], initial_spends=[10, 40], min_total_spend=50, max_total_spend=50, fix_props=True)
    assert len(pkg.adjustables) == 0
    adjustments.append(pkg)
    adjustments.append(at.SpendingAdjustment("Risk avoidance", 2020, "abs", 0.0, 200.0))
    adjustments.append(at.SpendingAdjustment("Harm reduction 1", 2020, "abs", 0.0, 200.0))
    measurables = at.MaximizeMeasurable("ch_all", [2020, np.inf])
    constraints = at.TotalSpendConstraint()  # Cap total spending in all years
    optimization = at.Optimization(name="default", adjustments=adjustments, measurables=measurables, constraints=constraints)  # Evaluate from 2020 to end of simulation

    (unoptimized_result, optimized_result) = run_optimization(P, optimization, instructions)

    for prog in alloc:
        print("%s - before=%.2f, after=%.2f" % (prog, unoptimized_result.model.program_instructions.alloc[prog].get(2020), optimized_result.model.program_instructions.alloc[prog].get(2020)))  # TODO - add time to alloc

    # The package should have the full $100 allocated to it, but split in the same 10:40 ratio i.e. $80 on treatment 2 and $20 on treatment 1
    # assert np.isclose(optimized_result.model.program_instructions.alloc["Treatment 2"].get(2020), 40)
    # assert np.isclose(optimized_result.model.program_instructions.alloc["Treatment 1"].get(2020), 10)


if __name__ == "__main__":
    test_standard()
    test_unresolvable()
    test_standard_mindeaths()
    test_delayed()
    test_multiyear_fixed()
    test_multiyear_relative()
    test_gradual()
    test_mixed_timing()
    test_parametric_paired()
    test_minmoney()
    test_minmoney_relative()
    test_minmoney_absolute()
    test_cascade_final_stage()
    test_cascade_multi_stage()
    test_cascade_conversions()
    test_package_fixed()
    test_package_variable()
    test_package_fixed_prop()
    test_package_all_fixed()
